# Standard Library
import datetime
import sys

# Third Party
from pydantic import BaseModel, conint, conlist, constr

# First Party
from resc_backend.db.model import DBfinding


class FindingBase(BaseModel):
    file_path: constr(max_length=500)
    line_number: conint(gt=-1)
    column_start: conint(gt=-1)
    column_end: conint(gt=-1)
    commit_id: constr(max_length=120)
    commit_message: str
    commit_timestamp: datetime.datetime
    author: constr(max_length=200)
    email: constr(max_length=100)
    event_sent_on: datetime.datetime | None
    rule_name: constr(max_length=400)


class FindingPatch(BaseModel):
    event_sent_on: datetime.datetime


class FindingCreate(FindingBase):
    repository_id: conint(gt=0)

    @classmethod
    def create_from_base_class(cls, base_object: FindingBase, repository_id: int):
        return cls(**(dict(base_object)), repository_id=repository_id)


class Finding(FindingBase):
    pass


class FindingRead(FindingCreate):
    id_: conint(gt=0)
    scan_ids: conlist(conint(gt=0), min_items=None, max_items=sys.maxsize) | None

    class Config:
        orm_mode = True

    @classmethod
    def create_from_db_entities(cls, db_finding: DBfinding, scan_ids: list[int]):
        return FindingRead(
            id_=db_finding.id_,
            file_path=db_finding.file_path,
            line_number=db_finding.line_number,
            column_start=db_finding.column_start,
            column_end=db_finding.column_end,
            commit_id=db_finding.commit_id,
            commit_message=db_finding.commit_message,
            commit_timestamp=db_finding.commit_timestamp,
            author=db_finding.author,
            email=db_finding.email,
            event_sent_on=db_finding.event_sent_on,
            rule_name=db_finding.rule_name,
            repository_id=db_finding.repository_id,
            scan_ids=scan_ids,
        )



# Third Party
from pydantic import BaseModel, conint, constr


class DateCountModel(BaseModel):
    date_lable: constr(max_length=100)
    finding_count: conint(gt=-1) = 0

    class Config:
        orm_mode = True



# Third Party
from sqlalchemy import (
    Column,
    Float,
    ForeignKey,
    Integer,
    String,
    Text,
    UniqueConstraint,
)

# First Party
from resc_backend.db.model import Base
from resc_backend.db.model.rule_allow_list import DBruleAllowList
from resc_backend.db.model.rule_pack import DBrulePack


class DBrule(Base):
    __tablename__ = "rules"
    id_ = Column("id", Integer, primary_key=True)
    rule_pack = Column(String(100), ForeignKey(DBrulePack.version), nullable=False)
    allow_list = Column(Integer, ForeignKey(DBruleAllowList.id_), nullable=True)
    rule_name = Column(String(400), nullable=False)
    description = Column(String(2000), nullable=True)
    entropy = Column(Float, nullable=True)
    secret_group = Column(Integer, nullable=True)
    regex = Column(Text, nullable=True)
    path = Column(Text, nullable=True)
    keywords = Column(Text, nullable=True)
    __table_args__ = (UniqueConstraint("rule_name", "rule_pack", name="unique_rule_name_per_rule_pack_version"),)

    def __init__(
        self,
        rule_pack: str,
        rule_name: str,
        description: str,
        allow_list: int = None,
        entropy: float = None,
        secret_group: str = None,
        regex: str = None,
        path: str = None,
        keywords: str = None,
    ):
        self.rule_pack = rule_pack
        self.allow_list = allow_list
        self.rule_name = rule_name
        self.description = description
        self.entropy = entropy
        self.secret_group = secret_group
        self.regex = regex
        self.path = path
        self.keywords = keywords

    @staticmethod
    def create_from_metadata(
        rule_pack: str,
        rule_name: str,
        description: str,
        entropy: float,
        secret_group: str,
        regex: str,
        path: str,
        keywords: str,
        allow_list: int,
    ):
        db_rule = DBrule(
            rule_pack=rule_pack,
            rule_name=rule_name,
            description=description,
            entropy=entropy,
            secret_group=secret_group,
            regex=regex,
            path=path,
            keywords=keywords,
            allow_list=allow_list,
        )
        return db_rule



# Standard Library

# Third Party
from sqlalchemy import func
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import (
    DBaudit,
    DBfinding,
    DBrepository,
    DBrule,
    DBruleTag,
    DBscan,
    DBscanFinding,
    DBtag,
    DBVcsInstance,
)
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.schema import (
    detailed_finding as detailed_finding_schema,
)
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType


def _query_join_if_multiple_rule_pack(query: Query, rule_pack_versions: list[str]) -> Query:
    """
    When we are working with rule pack it is sometimes necessary to use a join
    with a subquery (i.e. when there are 0 or > 1 rule pack selected).

    This takes care of the join of the sub query.

    Note that the filtering per rulepack is not applied here. We only search for the max
    base scan id for the specific set of rulepacks.

    Args:
        query (Query): Query to work on
        rule_pack_versions (List[str]): rule pack versions to filter on.

    Returns:
        Query: Updated query
    """
    if rule_pack_versions is not None and len(rule_pack_versions) == 1:
        return query

    subquery: Query = query.session.query(DBscan.repository_id, func.max(DBscan.id_).label("latest_base_scan_id"))

    # This contraint is not necessary, but it will make the table smaller.
    subquery = subquery.where(DBscan.is_latest == True)  # noqa: E712
    subquery = subquery.where(DBscan.scan_type == ScanType.BASE)
    if rule_pack_versions is not None and len(rule_pack_versions) > 0:
        subquery = subquery.where(DBscan.rule_pack.in_(rule_pack_versions))
    subquery = subquery.group_by(DBscan.repository_id)
    subquery = subquery.subquery()

    query = query.join(
        subquery,
        DBfinding.repository_id == subquery.c.repository_id,
    )
    query = query.where(DBscan.id_ >= subquery.c.latest_base_scan_id)

    return query


def _query_join_if_rule_tag(query: Query, rule_tags: list[str]) -> Query:
    """
    When we are working with a set of Rule tag with need to join a few extra tables.
    This filtering requires a few extra tables, hence if the field is None,
    we do not apply the join.

    Args:
        query (Query): query to join on.
        rule_tags (List[str] | None): tags to apply filter.

    Returns:
        Query: query with rule tag filtering applied
    """
    if rule_tags is None:
        return query

    subquery: Query = query.session.query(DBruleTag.rule_id)
    subquery = subquery.join(DBtag, DBruleTag.tag_id == DBtag.id_)
    subquery = subquery.where(DBtag.name.in_(rule_tags))
    subquery = subquery.group_by(DBruleTag.rule_id)
    subquery = subquery.subquery()

    query = query.join(
        DBrule,
        (DBrule.rule_name == DBfinding.rule_name) & (DBrule.rule_pack == DBscan.rule_pack),
    )
    query = query.where(DBrule.id_.in_(subquery))
    return query


def _query_apply_findings_filters(query: Query, findings_filter: FindingsFilter) -> Query:
    """
    Apply filtering on the query.
    This only applies filtering for:
        - rule_pack_versions
        - scan_ids
        - rule_names
        - start_date_time
        - event_sent
        - repository_name
        - vcs_providers
        - project_name
        - finding_statuses

    Args:
        query (Query): query to join on.
        findings_filter (FindingsFilter): Filters to apply.

    Returns:
        Query: query with filtering applied
    """
    if findings_filter.rule_pack_versions is not None:
        query = (
            query.where(DBscan.rule_pack == findings_filter.rule_pack_versions[0])
            if len(findings_filter.rule_pack_versions) == 1
            else query.where(DBscan.rule_pack.in_(findings_filter.rule_pack_versions))
        )

    if findings_filter.scan_ids:
        query = query.where(
            DBscan.id_.in_(findings_filter.scan_ids),
        )
    else:
        query = query.where(DBscan.is_latest == True)  # noqa: E712

    if findings_filter.rule_names:
        query = query.where(DBfinding.rule_name.in_(findings_filter.rule_names))

    if findings_filter.start_date_time:
        query = query.where(DBscan.timestamp >= findings_filter.start_date_time)

    if findings_filter.end_date_time:
        query = query.where(DBscan.timestamp <= findings_filter.end_date_time)

    if findings_filter.event_sent is not None:
        if findings_filter.event_sent:
            query = query.where(DBfinding.event_sent_on != None)  # noqa: E711
        else:
            query = query.where(DBfinding.event_sent_on == None)  # noqa: E711

    if findings_filter.repository_name:
        query = query.where(DBrepository.repository_name == findings_filter.repository_name)

    if findings_filter.vcs_providers and findings_filter.vcs_providers is not None:
        query = query.where(DBVcsInstance.provider_type.in_(findings_filter.vcs_providers))

    if findings_filter.project_name:
        query = query.where(DBrepository.project_key == findings_filter.project_name)

    if findings_filter.finding_statuses:
        if FindingStatus.NOT_ANALYZED.value in findings_filter.finding_statuses:
            query = query.where(
                (DBaudit.status.in_(findings_filter.finding_statuses)) | (DBaudit.status == None)  # noqa: E711
            )
        else:
            query = query.where(DBaudit.status.in_(findings_filter.finding_statuses))
    return query


def get_detailed_findings(
    db_connection: Session,
    findings_filter: FindingsFilter,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
) -> list[detailed_finding_schema.DetailedFindingRead]:
    """
    Retrieve all detailed findings objects matching the provided FindingsFilter
    :param findings_filter:
        Object of type FindingsFilter, only DetailedFindingRead objects matching the attributes in this filter will be
            fetched
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DetailedFindingRead]
        The output will contain a list of DetailedFindingRead objects,
        or an empty list if no finding was found for the given findings_filter
    """

    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    query = db_connection.query(
        DBfinding.id_,
        DBfinding.file_path,
        DBfinding.line_number,
        DBfinding.column_start,
        DBfinding.column_end,
        DBfinding.commit_id,
        DBfinding.commit_message,
        DBfinding.commit_timestamp,
        DBfinding.author,
        DBfinding.email,
        DBaudit.status,
        DBaudit.comment,
        DBfinding.rule_name,
        DBscan.rule_pack,
        DBfinding.event_sent_on,
        DBscan.timestamp,
        DBscan.id_.label("scan_id"),
        DBscan.last_scanned_commit,
        DBVcsInstance.provider_type.label("vcs_provider"),
        DBrepository.project_key,
        DBrepository.repository_name,
        DBrepository.repository_url,
    )
    # Prepare the joins, then we do the filtering.
    # There are no benefits in doing filtering on Where rather than On as we are doing Inner joins.
    # Source: https://stackoverflow.com/questions/2509987/which-sql-query-is-faster-filter-on-join-criteria-or-where-clause
    # The goal here is to avoid sub queries.
    query = query.join(DBscanFinding, DBfinding.id_ == DBscanFinding.finding_id)
    query = query.join(DBrepository, DBfinding.repository_id == DBrepository.id_)
    query = query.join(DBVcsInstance, DBrepository.vcs_instance == DBVcsInstance.id_)
    query = query.join(DBscan, DBscanFinding.scan_id == DBscan.id_)
    query = _query_join_if_multiple_rule_pack(query, findings_filter.rule_pack_versions)
    query = _query_join_if_rule_tag(query, findings_filter.rule_tags)
    query = query.join(
        DBaudit,
        (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
        isouter=True,
    )
    query = _query_apply_findings_filters(query, findings_filter)
    query = query.order_by(DBfinding.id_)
    query = query.offset(skip).limit(limit_val)
    findings: list[detailed_finding_schema.DetailedFindingRead] = query.all()

    return findings


def get_detailed_findings_count(db_connection: Session, findings_filter: FindingsFilter) -> int:
    """
    Retrieve count of detailed findings objects matching the provided FindingsFilter
    :param findings_filter:
        Object of type FindingsFilter, only DetailedFindingRead objects matching the attributes in this filter will be
            fetched
    :param db_connection:
        Session of the database connection
    :return: total_count
        count of findings
    """

    query = db_connection.query(func.count(DBfinding.id_))
    # Prepare the joins, then we do the filtering.
    # There are no benefits in doing filtering on Where rather than On as we are doing Inner joins.
    # Source: https://stackoverflow.com/questions/2509987/which-sql-query-is-faster-filter-on-join-criteria-or-where-clause
    # The goal here is to avoid sub queries.
    query = query.join(DBscanFinding, DBfinding.id_ == DBscanFinding.finding_id)
    query = query.join(DBrepository, DBfinding.repository_id == DBrepository.id_)
    query = query.join(DBVcsInstance, DBrepository.vcs_instance == DBVcsInstance.id_)
    query = query.join(DBscan, DBscanFinding.scan_id == DBscan.id_)
    query = _query_join_if_multiple_rule_pack(query, findings_filter.rule_pack_versions)
    query = _query_join_if_rule_tag(query, findings_filter.rule_tags)
    query = query.join(
        DBaudit,
        (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
        isouter=True,
    )
    query = _query_apply_findings_filters(query, findings_filter)
    findings_count = query.scalar()
    return findings_count


def get_detailed_finding(db_connection: Session, finding_id: int) -> detailed_finding_schema.DetailedFindingRead:
    """
    Retrieve a detailed finding objects matching the provided finding_id
    :param db_connection:
        Session of the database connection
    :param finding_id:
        ID of the finding object for which a DetailedFinding is to be fetched
    :return: DetailedFindingRead
        The output will contain an object of type DetailedFindingRead,
            or a null object finding was found for the given finding_id
    """
    query = (
        db_connection.query(
            DBfinding.id_,
            DBfinding.file_path,
            DBfinding.line_number,
            DBfinding.column_start,
            DBfinding.column_end,
            DBfinding.commit_id,
            DBfinding.commit_message,
            DBfinding.commit_timestamp,
            DBfinding.author,
            DBfinding.email,
            DBaudit.status,
            DBaudit.comment,
            DBfinding.rule_name,
            DBscan.rule_pack,
            DBscan.timestamp,
            DBscan.id_.label("scan_id"),
            DBscan.last_scanned_commit,
            DBVcsInstance.provider_type.label("vcs_provider"),
            DBrepository.project_key,
            DBrepository.repository_name,
            DBrepository.repository_url,
        )
        .join(DBscanFinding, DBfinding.id_ == DBscanFinding.finding_id)
        .join(
            DBscan,
            (DBscan.id_ == DBscanFinding.scan_id) & (DBscan.is_latest == True),  # noqa: E712
        )
        .join(DBrepository, DBrepository.id_ == DBscan.repository_id)
        .join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        .join(
            DBaudit,
            (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
            isouter=True,
        )
        .where(DBfinding.id_ == finding_id)
        .order_by(DBscan.id_.desc())
    )
    finding = query.first()
    return finding



# Standard Library

# Third Party
from pydantic import BaseModel, conint, constr


class RuleBase(BaseModel):
    rule_name: constr(min_length=1, max_length=400)
    description: constr(max_length=2000) | None = None
    entropy: float | None = None
    secret_group: int | None = None
    regex: str | None = None
    path: str | None = None
    keywords: str | None = None


class RuleCreate(RuleBase):
    rule_pack: constr(regex=r"^(\d+\.)?(\d+\.)?(\*|\d+)$")
    allow_list: conint(gt=0) | None = None

    @classmethod
    def create_from_base_class(cls, base_object: RuleBase, rule_pack: str, allow_list=int):
        return cls(**(dict(base_object)), rule_pack=rule_pack, allow_list=allow_list)


class Rule(RuleBase):
    pass


class RuleRead(RuleCreate):
    id_: conint(gt=0)

    class Config:
        orm_mode = True



# Standard Library
import logging

# Third Party
import requests

# First Party
from resc_backend.constants import RWS_ROUTE_SCANS, RWS_VERSION_PREFIX
from resc_backend.resc_web_service.schema.scan import ScanCreate

logger = logging.getLogger(__name__)


def create_scan(url: str, scan: ScanCreate):
    api_url = f"{url}{RWS_VERSION_PREFIX}{RWS_ROUTE_SCANS}"
    response = requests.post(api_url, data=scan.json(), proxies={"http": "", "https": ""}, timeout=10)
    return response



# Standard Library
import json
import logging

# Third Party
import requests

# First Party
from resc_backend.constants import (
    RWS_ROUTE_FINDINGS,
    RWS_ROUTE_SCANS,
    RWS_VERSION_PREFIX,
)
from resc_backend.resc_web_service.schema.finding import FindingCreate

logger = logging.getLogger(__name__)


def create_findings(url: str, findings: list[FindingCreate]) -> requests.Response:
    api_url = f"{url}{RWS_VERSION_PREFIX}{RWS_ROUTE_FINDINGS}"

    findings_json = []
    for finding in findings:
        findings_json.append(json.loads(finding.json()))

    response = requests.post(api_url, json=findings_json, proxies={"http": "", "https": ""}, timeout=10)
    return response


def create_findings_with_scan_id(url: str, findings: list[FindingCreate], scan_id: int) -> requests.Response:
    api_url = f"{url}{RWS_VERSION_PREFIX}{RWS_ROUTE_SCANS}/{scan_id}{RWS_ROUTE_FINDINGS}"

    findings_json = []
    for finding in findings:
        findings_json.append(json.loads(finding.json()))

    response = requests.post(api_url, json=findings_json, proxies={"http": "", "https": ""}, timeout=10)
    return response



# Third Party
from sqlalchemy import Column, Integer, String, UniqueConstraint

# First Party
from resc_backend.db.model import Base


class DBtag(Base):
    __tablename__ = "tag"
    id_ = Column("id", Integer, primary_key=True)
    name = Column(String(200), nullable=False)
    __table_args__ = (UniqueConstraint("name", name="unique_tag"),)

    def __init__(self, name: str):
        self.name = name



# Third Party
from pydantic import BaseModel, conint


class PersonalAuditMetrics(BaseModel):
    today: conint(gt=-1) = 0
    current_week: conint(gt=-1) = 0
    last_week: conint(gt=-1) = 0
    current_month: conint(gt=-1) = 0
    current_year: conint(gt=-1) = 0
    forever: conint(gt=-1) = 0
    rank_current_week: conint(gt=-1) = 0



# Standard Library
from datetime import datetime, UTC

# Third Party
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String

# First Party
from resc_backend.db.model import Base
from resc_backend.db.model.rule_allow_list import DBruleAllowList


class DBrulePack(Base):
    __tablename__ = "rule_pack"
    version = Column("version", String(100), primary_key=True)
    global_allow_list = Column(Integer, ForeignKey(DBruleAllowList.id_), nullable=True)
    active = Column(Boolean, nullable=False, default=False)
    created = Column(DateTime, nullable=False, default=datetime.now(UTC))

    def __init__(
        self,
        version: str,
        global_allow_list: int = None,
        active: bool = False,
        created: datetime = datetime.now(UTC),
    ):
        self.version = version
        self.global_allow_list = global_allow_list
        self.active = active
        self.created = created

    @staticmethod
    def create_from_metadata(version: str, global_allow_list: int, active: bool, created: datetime):
        db_rule_pack = DBrulePack(
            version=version,
            global_allow_list=global_allow_list,
            active=active,
            created=created,
        )
        return db_rule_pack



# Standard Library
from enum import Enum
from typing import TypedDict


class StatusStats(TypedDict):
    true_positive: int
    false_positive: int
    not_analyzed: int
    not_accessible: int
    clarification_required: int
    total_findings_count: int


class FindingStatus(str, Enum):
    NOT_ANALYZED = "NOT_ANALYZED"
    NOT_ACCESSIBLE = "NOT_ACCESSIBLE"
    CLARIFICATION_REQUIRED = "CLARIFICATION_REQUIRED"
    FALSE_POSITIVE = "FALSE_POSITIVE"
    TRUE_POSITIVE = "TRUE_POSITIVE"
    OUTDATED = "OUTDATED"

    @classmethod
    def init_statistics(cls) -> StatusStats:
        return {
            "true_positive": 0,
            "false_positive": 0,
            "not_analyzed": 0,
            "not_accessible": 0,
            "clarification_required": 0,
            "outdated": 0,
            "total_findings_count": 0,
        }



# Third Party
from sqlalchemy import Column, Integer, String, Text

# First Party
from resc_backend.db.model import Base


class DBruleAllowList(Base):
    __tablename__ = "rule_allow_list"
    id_ = Column("id", Integer, primary_key=True)
    description = Column(String(2000), nullable=True)
    regexes = Column(Text, nullable=True)
    paths = Column(Text, nullable=True)
    commits = Column(Text, nullable=True)
    stop_words = Column(Text, nullable=True)

    def __init__(
        self,
        description: str,
        regexes: str = None,
        paths: str = None,
        commits: str = None,
        stop_words: str = None,
    ):
        self.description = description
        self.regexes = regexes
        self.paths = paths
        self.commits = commits
        self.stop_words = stop_words

    @staticmethod
    def create_from_metadata(description: str, regexes: str, paths: str, commits: str, stop_words: str):
        db_rule_allow_list = DBruleAllowList(
            description=description,
            regexes=regexes,
            paths=paths,
            commits=commits,
            stop_words=stop_words,
        )
        return db_rule_allow_list



# Standard Library
import logging
import os
import struct

# Third Party
from azure.identity import DefaultAzureCredential
from sqlalchemy import create_engine, event

basedir = os.path.abspath(os.path.dirname(__file__))
logger = logging.getLogger(__name__)


def create_azure_engine(connection_string: str, echo_queries: bool, pool_size: int, max_overflow: int):
    """
        Create a SQLAlchemy Engine with added azure token injection
    :param connection_string:
        the connection string to the database
    :param echo_queries:
        bool, echo queries as defined in SQLAlchemy
    :param pool_size:
        integer, pool size as defined in SQLAlchemy
    :param max_overflow:
        integer, max overflow as defined in SQLAlchemy
    :return: SQLAlchemyEngine
        The output will contain the engine
    """

    managed_identity_client_id = os.environ.get("MANAGED_IDENTITY_CLIENT_ID")
    if not managed_identity_client_id:
        logger.error("Missing environment variable MANAGED_IDENTITY_CLIENT_ID")
    credential = DefaultAzureCredential(
        managed_identity_client_id=managed_identity_client_id,
        logging_enable=True,
        logging_body=True,
    )
    sql_copt_ss_access_token = 1256

    logger.info("Using provided environment variable to connect to the Database with azure token auth")
    engine = create_engine(
        connection_string,
        echo=echo_queries,
        pool_size=pool_size,
        max_overflow=max_overflow,
    )

    @event.listens_for(engine, "do_connect")
    def provide_token(dialect, conn_rec, cargs, cparams):
        token = credential.get_token("https://database.windows.net/.default").token.encode("UTF-16-LE")
        token_struct = struct.pack(f"<I{len(token)}s", len(token), token)
        cparams["attrs_before"] = {sql_copt_ss_access_token: token_struct}

    return engine



# Standard Library

# Third Party
from fastapi import APIRouter, status
from fastapi_cache.decorator import cache

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_VCS_INSTANCE,
    COMMON_TAG,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    REDIS_CACHE_EXPIRE,
    RWS_ROUTE_AUTH_CHECK,
    RWS_ROUTE_SUPPORTED_VCS_PROVIDERS,
)
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(tags=[COMMON_TAG])


@router.get(
    f"{RWS_ROUTE_SUPPORTED_VCS_PROVIDERS}",
    response_model=list[str],
    summary="Get supported vcs-providers",
    description="Retrieve the supported vcs-providers, example: Bitbucket, AzureDevOps, Github etc",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve the supported vcs-providers"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_VCS_INSTANCE, expire=REDIS_CACHE_EXPIRE)
def get_supported_vcs_providers() -> list[str]:
    """
        Retrieve all supported vcs providers
    :return: List[str]
        The output will contain a list of strings of unique vcs providers
    """
    supported_vcs = [vcs for vcs in VCSProviders if vcs]
    return supported_vcs


@router.get(
    f"{RWS_ROUTE_AUTH_CHECK}",
    summary="Authorization check",
    description="The output returns 200 OK if auth check is successful else returns 403 Forbidden",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Validate authorization check from the access-token"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def auth_check():
    """
        Validates authorization check from the access token
    :return: str
        The output will contain 200 OK if auth check is successful else it will return 403 Forbidden
    """
    return {"message": "OK"}



# Standard Library
from datetime import datetime, UTC

# Third Party
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
    Text,
    UniqueConstraint,
)

# First Party
from resc_backend.db.model import Base


class DBfinding(Base):
    __tablename__ = "finding"
    id_ = Column("id", Integer, primary_key=True)
    repository_id = Column(Integer, ForeignKey("repository.id"), nullable=False)
    rule_name = Column(String(400), nullable=False)
    file_path = Column(String(500), nullable=False)
    line_number = Column(Integer, nullable=False)
    column_start = Column(Integer, nullable=False, default=0)
    column_end = Column(Integer, nullable=False, default=0)
    commit_id = Column(String(120))
    commit_message = Column(Text)
    commit_timestamp = Column(DateTime, default=datetime.now(UTC).isoformat())
    author = Column(String(200))
    email = Column(String(100))
    event_sent_on = Column(DateTime, nullable=True)

    __table_args__ = (
        UniqueConstraint(
            "commit_id",
            "repository_id",
            "rule_name",
            "file_path",
            "line_number",
            "column_start",
            "column_end",
            name="uc_finding_per_repository",
        ),
    )

    def __init__(
        self,
        rule_name,
        file_path,
        line_number,
        commit_id,
        commit_message,
        commit_timestamp,
        author,
        email,
        event_sent_on,
        repository_id,
        column_start,
        column_end,
    ):
        self.email = email
        self.author = author
        self.commit_timestamp = commit_timestamp
        self.commit_message = commit_message
        self.commit_id = commit_id
        self.line_number = line_number
        self.file_path = file_path
        self.rule_name = rule_name
        self.event_sent_on = event_sent_on
        self.repository_id = repository_id
        self.column_start = column_start
        self.column_end = column_end

    @staticmethod
    def create_from_finding(finding):
        db_finding = DBfinding(
            rule_name=finding.rule_name,
            file_path=finding.file_path,
            line_number=finding.line_number,
            email=finding.email,
            commit_id=finding.commit_id,
            commit_message=finding.commit_message,
            commit_timestamp=finding.commit_timestamp,
            author=finding.author,
            event_sent_on=finding.event_sent_on,
            repository_id=finding.repository_id,
            column_start=finding.column_start,
            column_end=finding.column_end,
        )
        return db_finding



# Standard Library

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, status
from fastapi_cache.decorator import cache

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    CACHE_NAMESPACE_REPOSITORY,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    REDIS_CACHE_EXPIRE,
    REPOSITORIES_TAG,
    RWS_ROUTE_DISTINCT_PROJECTS,
    RWS_ROUTE_DISTINCT_REPOSITORIES,
    RWS_ROUTE_FINDINGS_METADATA,
    RWS_ROUTE_LAST_SCAN,
    RWS_ROUTE_REPOSITORIES,
    RWS_ROUTE_SCANS,
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import repository as repository_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.helpers.resc_swagger_models import Model404
from resc_backend.resc_web_service.schema import repository as repository_schema
from resc_backend.resc_web_service.schema import (
    repository_enriched as repository_enriched_schema,
)
from resc_backend.resc_web_service.schema import scan as scan_schema
from resc_backend.resc_web_service.schema.finding_count_model import FindingCountModel
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(prefix=f"{RWS_ROUTE_REPOSITORIES}", tags=[REPOSITORIES_TAG])


@router.get(
    "",
    response_model=PaginationModel[repository_schema.RepositoryRead],
    summary="Get repositories",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the repositories"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_all_repositories(
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    vcsproviders: list[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
    projectfilter: str | None = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    repositoryfilter: str | None = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[repository_schema.RepositoryRead]:
    """
        Retrieve all repository objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **vcsproviders**: Optional, filter on supported vcs provider types
    - **projectfilter**: Optional, filter on project name. It is used as a string contains filter
    - **repositoryfilter**: Optional, filter on repository name. It is used as a string contains filter
    - **return**: [RepositoryRead]
        The output will contain a PaginationModel containing the list of RepositoryRead type objects,
        or an empty list if no repository
    """

    repositories = repository_crud.get_repositories(
        db_connection,
        skip=skip,
        limit=limit,
        vcs_providers=vcsproviders,
        project_filter=projectfilter,
        repository_filter=repositoryfilter,
    )

    total_repositories = repository_crud.get_repositories_count(
        db_connection,
        vcs_providers=vcsproviders,
        project_filter=projectfilter,
        repository_filter=repositoryfilter,
    )

    return PaginationModel[repository_schema.RepositoryRead](
        data=repositories, total=total_repositories, limit=limit, skip=skip
    )


@router.post(
    "",
    response_model=repository_schema.RepositoryRead,
    summary="Create a repository",
    status_code=status.HTTP_201_CREATED,
    responses={
        201: {"description": "Create a new repository"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def create_repository(
    repository: repository_schema.RepositoryCreate,
    db_connection: Session = Depends(get_db_connection),
):
    """
        Create a repository with all the information

    - **db_connection**: Session of the database connection
    - **project_key**: each repository must have a project name or key
    - **repository_id**: repository id
    - **repository_name**: repository name
    - **repository_url**: repository url
    - **vcs_instance**: vcs instance id
    """
    repository = repository_crud.create_repository_if_not_exists(db_connection=db_connection, repository=repository)

    # Clear cache related to repository
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_REPOSITORY)
    return repository


@router.get(
    "/{repository_id}",
    response_model=repository_schema.RepositoryRead,
    summary="Fetch a repository by ID",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve repository <repository_id>"},
        404: {"model": Model404, "description": "Repository <repository_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def read_repository(repository_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Read a repository by ID

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository for which details need to be fetched
    """
    db_repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if db_repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")
    return db_repository


@router.put(
    "/{repository_id}",
    response_model=repository_schema.RepositoryRead,
    summary="Update an existing repository",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Update repository <repository_id>"},
        404: {"model": Model404, "description": "Repository <repository_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def update_repository(
    repository_id: int,
    repository: repository_schema.RepositoryCreate,
    db_connection: Session = Depends(get_db_connection),
):
    """
        Update an existing repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository
    - **project_key**: project name that needs to be updated
    - **repository_id**: repository id that needs to be updated
    - **repository_name**: repository name that needs to be updated
    - **repository_url**: repository url that needs to be updated
    - **vcs_instance**: vcs instance id that needs to be updated
    """
    db_repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if db_repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")

    updated_repository = repository_crud.update_repository(
        db_connection=db_connection, repository_id=repository_id, repository=repository
    )

    # Clear cache related to repository
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_REPOSITORY)
    return updated_repository


@router.delete(
    "/{repository_id}",
    summary="Delete a repository",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Delete repository <repository_id>"},
        404: {"model": Model404, "description": "Repository <repository_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def delete_repository(repository_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Delete a repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository to delete
    - **return**: The output will contain a success or error message based on the success of the deletion
    """
    db_repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if db_repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")
    repository_crud.delete_repository(db_connection, repository_id=repository_id, delete_related=True)

    # Clear cache related to repository
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_REPOSITORY)
    return {"ok": True}


@router.get(
    f"{RWS_ROUTE_DISTINCT_PROJECTS}/",
    response_model=list[str],
    summary="Get all unique project names",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the unique project-names"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_REPOSITORY, expire=REDIS_CACHE_EXPIRE)
def get_distinct_projects(
    vcsproviders: list[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
    repositoryfilter: str | None = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    onlyifhasfindings: bool = Query(default=False),
    db_connection: Session = Depends(get_db_connection),
) -> list[str]:
    """
        Retrieve all unique project names

    - **db_connection**: Session of the database connection
    - **vcsproviders**: Optional, filter on supported vcs provider types
    - **repositoryfilter**: Optional, filter on repository name. It is used as a string contains filter
    - **onlyifhasfindings**: Optional, filter all projects those have findings
    - **return**: List[str]
        The output will contain a list of unique projects
    """

    distinct_projects = repository_crud.get_distinct_projects(
        db_connection,
        vcs_providers=vcsproviders,
        repository_filter=repositoryfilter,
        only_if_has_findings=onlyifhasfindings,
    )
    projects = [project.project_key for project in distinct_projects]
    return projects


@router.get(
    f"{RWS_ROUTE_DISTINCT_REPOSITORIES}/",
    response_model=list[str],
    summary="Get all unique repository names",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the unique repository names"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_REPOSITORY, expire=REDIS_CACHE_EXPIRE)
def get_distinct_repositories(
    vcsproviders: list[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
    projectname: str | None = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    onlyifhasfindings: bool = Query(default=False),
    db_connection: Session = Depends(get_db_connection),
) -> list[str]:
    """
        Retrieve all unique repository names

    - **db_connection**: Session of the database connection
    - **vcsproviders**: Optional, filter of supported vcs provider types
    - **projectname**: Optional, filter on project name. It is used as a full string match filter
    - **onlyifhasfindings**: Optional, filter all repositories that have findings
    - **return**: List[str]
        The output will contain a list of unique repositories
    """

    distinct_repositories = repository_crud.get_distinct_repositories(
        db_connection,
        vcs_providers=vcsproviders,
        project_name=projectname,
        only_if_has_findings=onlyifhasfindings,
    )
    repositories = [repo.repository_name for repo in distinct_repositories]
    return repositories


@router.get(
    "/{repository_id}" f"{RWS_ROUTE_FINDINGS_METADATA}",
    response_model=FindingCountModel[repository_schema.RepositoryRead],
    summary="Get findings metadata for a repository",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve findings metadata for repository <repository_id>"},
        404: {"model": Model404, "description": "Repository <repository_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_findings_metadata_for_repository(
    repository_id: int, db_connection: Session = Depends(get_db_connection)
) -> FindingCountModel[repository_schema.RepositoryRead]:
    """
        Retrieve findings metadata for a repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the repository object for which findings metadata to be retrieved
    - **return**: RepositoryRead, findings count per status
        The output will contain a RepositoryRead type object along with findings count per status,
        or empty if no scan was found
    """
    repository = repository_crud.get_repository(db_connection, repository_id=repository_id)
    if repository is None:
        raise HTTPException(status_code=404, detail="Repository not found")

    findings_meta_data = repository_crud.get_findings_metadata_by_repository_id(
        db_connection, repository_ids=[repository_id]
    )

    return FindingCountModel[repository_schema.RepositoryRead](
        data=repository,
        true_positive=findings_meta_data[repository_id]["true_positive"],
        false_positive=findings_meta_data[repository_id]["false_positive"],
        not_analyzed=findings_meta_data[repository_id]["not_analyzed"],
        not_accessible=findings_meta_data[repository_id]["not_accessible"],
        clarification_required=findings_meta_data[repository_id]["clarification_required"],
        outdated=findings_meta_data[repository_id]["outdated"],
        total_findings_count=findings_meta_data[repository_id]["total_findings_count"],
    )


@router.get(
    f"{RWS_ROUTE_FINDINGS_METADATA}/",
    response_model=PaginationModel[repository_enriched_schema.RepositoryEnrichedRead],
    summary="Get all repositories with findings metadata",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the findings metadata for all the repositories"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_all_repositories_with_findings_metadata(
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    vcsproviders: list[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
    projectfilter: str | None = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    repositoryfilter: str | None = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    onlyifhasfindings: bool = Query(default=False),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[repository_enriched_schema.RepositoryEnrichedRead]:
    """
        Retrieve all repository objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **vcsproviders**: Optional, filter on supported vcs provider types
    - **projectfilter**: Optional, filter on project name. It is used as a string contains filter
    - **repositoryfilter**: Optional, filter on repository name. It is used as a string contains filter
    - **onlyifhasfindings**: Optional, filter all repositories those have findings
    - **return**: [RepositoryEnrichedRead]
        The output will contain a PaginationModel containing the list of RepositoryEnrichedRead type objects,
        or an empty list if no repository
    """

    repositories = repository_crud.get_repositories(
        db_connection,
        skip=skip,
        limit=limit,
        vcs_providers=vcsproviders,
        project_filter=projectfilter,
        repository_filter=repositoryfilter,
        only_if_has_findings=onlyifhasfindings,
    )

    total_repositories = repository_crud.get_repositories_count(
        db_connection,
        vcs_providers=vcsproviders,
        project_filter=projectfilter,
        repository_filter=repositoryfilter,
        only_if_has_findings=onlyifhasfindings,
    )
    repository_list = []
    repo_ids = [repo.id_ for repo in repositories]
    repo_findings_meta_data = repository_crud.get_findings_metadata_by_repository_id(
        db_connection, repository_ids=repo_ids
    )
    for repo in repositories:
        enriched_repository = repository_enriched_schema.RepositoryEnrichedRead(
            id_=repo.id_,
            project_key=repo.project_key,
            repository_id=repo.repository_id,
            repository_name=repo.repository_name,
            repository_url=repo.repository_url,
            vcs_provider=repo.provider_type,
            last_scan_id=repo.last_scan_id,
            last_scan_timestamp=repo.last_scan_timestamp,
            true_positive=repo_findings_meta_data[repo.id_]["true_positive"],
            false_positive=repo_findings_meta_data[repo.id_]["false_positive"],
            not_analyzed=repo_findings_meta_data[repo.id_]["not_analyzed"],
            not_accessible=repo_findings_meta_data[repo.id_]["not_accessible"],
            clarification_required=repo_findings_meta_data[repo.id_]["clarification_required"],
            outdated=repo_findings_meta_data[repo.id_]["outdated"],
            total_findings_count=repo_findings_meta_data[repo.id_]["total_findings_count"],
        )
        repository_list.append(enriched_repository)

    return PaginationModel[repository_enriched_schema.RepositoryEnrichedRead](
        data=repository_list, total=total_repositories, limit=limit, skip=skip
    )


@router.get(
    "/{repository_id}" f"{RWS_ROUTE_LAST_SCAN}",
    response_model=scan_schema.ScanRead,
    summary="Get latest scan for repository",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve the latest scan related to a repository"},
        404: {"description": "Scan not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_last_scan_for_repository(
    repository_id: int, db_connection: Session = Depends(get_db_connection)
) -> scan_schema.ScanRead:
    """
        Retrieve the latest scan object related to a repository

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the parent repository object for which scan objects to be retrieved
    - **return**: ScanRead
        The output will contain a ScanRead type object,
        or empty if no scan was found
    """
    last_scan = scan_crud.get_latest_scan_for_repository(db_connection, repository_id=repository_id)
    if last_scan is None:
        raise HTTPException(status_code=404, detail="Scan not found")
    return last_scan


@router.get(
    "/{repository_id}" f"{RWS_ROUTE_SCANS}",
    summary="Get scans for repository",
    response_model=PaginationModel[scan_schema.ScanRead],
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the scans related to a repository"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_scans_for_repository(
    repository_id: int,
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[scan_schema.ScanRead]:
    """
        Retrieve all scan objects related to a repository paginated

    - **db_connection**: Session of the database connection
    - **repository_id**: ID of the parent repository object for which scan objects to be retrieved
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [ScanRead]
        The output will contain a PaginationModel containing the list of ScanRead type objects,
        or an empty list if no scan was found
    """
    scans = scan_crud.get_scans(db_connection, skip=skip, limit=limit, repository_id=repository_id)

    total_scans = scan_crud.get_scans_count(db_connection, repository_id=repository_id)

    return PaginationModel[scan_schema.ScanRead](data=scans, total=total_scans, limit=limit, skip=skip)



# Standard Library
from datetime import datetime

# Third Party
from pydantic import validator
from pydantic.dataclasses import dataclass

# First Party
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


@dataclass
class FindingsFilter:
    vcs_providers: list[VCSProviders] = None
    finding_statuses: list[FindingStatus] = None
    rule_names: list[str] = None
    rule_tags: list[str] = None
    project_name: str = None
    repository_name: str = None
    scan_ids: list[int] = None
    start_date_time: datetime = None
    end_date_time: datetime = None
    event_sent: bool = None
    rule_pack_versions: list[str] = None

    @validator("end_date_time")
    @classmethod
    def date_range_check(cls, end_date_time: datetime, values: dict):
        if end_date_time and values["start_date_time"]:
            if values["start_date_time"] >= end_date_time:
                raise ValueError("the start of the date range needs to be prior to the end of it.")

        return end_date_time



# Standard Library
import logging
import os
import re

# Third Party
import tomlkit
from fastapi import File, HTTPException
from tomlkit import aot, comment, document, nl, table
from tomlkit.items import String, StringType

# First Party
from resc_backend.constants import (
    ALLOWED_EXTENSION,
    TEMP_RULE_FILE,
    TOML_CUSTOM_DELIMITER,
)
from resc_backend.resc_web_service.schema.rule import Rule
from resc_backend.resc_web_service.schema.rule_allow_list import RuleAllowList

FILE_NAME_REGEX = r"^[a-zA-Z0-9-_]+$"

logger = logging.getLogger(__name__)


def create_allow_list_dictionary(allow_list: RuleAllowList) -> dict:
    """
        Create a dictionary for allow list when an RuleAllowList object is supplied
    :param allow_list:
        RuleAllowList object
    :return: AllowList dictionary
        The output will contain a dictionary of AllowList
    """
    allow_list_dict = {}
    if allow_list:
        if allow_list.description:
            allow_list_dict["description"] = allow_list.description
        if allow_list.regexes:
            allow_list_dict["regexes"] = allow_list.regexes
        if allow_list.paths:
            allow_list_dict["paths"] = allow_list.paths
        if allow_list.commits:
            allow_list_dict["commits"] = allow_list.commits
        if allow_list.stop_words:
            allow_list_dict["stop_words"] = allow_list.stop_words
    return allow_list_dict


def create_rule_dictionary(rule: Rule, allow_list_dict: dict, tags: str) -> dict:
    """
        Create a dictionary for rule when Rule object and RuleAllowList dict are supplied
    :param rule:
        Rule object
    :param allow_list_dict:
        Allow list dictionary
    :param tags:
        String of tags of the rule
    :return: Rule dictionary
        The output will contain a dictionary of Rule
    """
    rule_dict = {}
    if rule.rule_name:
        rule_dict["id"] = rule.rule_name
        rule_dict["description"] = rule.rule_name
    if tags:
        rule_dict["tags"] = tags
    if rule.entropy:
        rule_dict["entropy"] = rule.entropy
    if rule.secret_group:
        rule_dict["secret_group"] = rule.secret_group
    if rule.regex:
        rule_dict["regex"] = rule.regex
    if rule.path:
        rule_dict["path"] = rule.path
    if rule.keywords:
        rule_dict["keywords"] = rule.keywords
    if allow_list_dict:
        rule_dict["allow_list"] = allow_list_dict
    return rule_dict


def create_toml_dictionary(
    rule_pack_version: str,
    rules: list[str],
    global_allow_list: list[str],
    rule_tag_names,
) -> dict:
    """
        Create a dictionary for gitleaks toml rule for specified rule pack version, rules and global allow list
    :param rule_pack_version:
        Rule pack version
    :param rules:
        Rule list
    :param global_allow_list:
        Global Allow list
    :param rule_tag_names:
        List of rule names and tags
    :return: toml dictionary
        The output will contain a dictionary for gitleaks toml rule
    """
    rule_list = []
    for rule in rules:
        allow_list_dict = create_allow_list_dictionary(rule)
        tags_list = [x.name for x in rule_tag_names if x.rule_name == rule.rule_name]
        tags = None
        if len(tags_list) >= 1:
            tags = ",".join(tags_list)
        rule_dict = create_rule_dictionary(rule, allow_list_dict, tags)
        rule_list.append(rule_dict)

    global_allow_list_dict = create_allow_list_dictionary(global_allow_list)

    rule_toml_dict = {"title": "gitleaks config"}
    if rule_pack_version:
        rule_toml_dict["version"] = rule_pack_version
    if rule_list:
        rule_toml_dict["rules"] = rule_list
    if global_allow_list_dict:
        rule_toml_dict["allowlist"] = global_allow_list_dict
    return rule_toml_dict


def get_mapped_global_allow_list_obj(toml_rule_dictionary: dict) -> RuleAllowList:
    """
        Get global allow list object from toml rule dictionary
    :param toml_rule_dictionary:
        TOML rule dictionary
    :return: RuleAllowList
        The output will contain an object of RuleAllowList
    """
    global_allow_list_obj = None
    if "allowlist" in toml_rule_dictionary:
        global_allow_list = toml_rule_dictionary.get("allowlist")
        global_allow_list_obj = map_dictionary_to_rule_allow_list_object(global_allow_list)
    else:
        logger.info("No global allow list is present in the toml file!")
    return global_allow_list_obj


def map_dictionary_to_rule_allow_list_object(
    allow_list_dictionary: dict,
) -> RuleAllowList:
    """
        Convert allow list dictionary to RuleAllowList object
    :param allow_list_dictionary:
        AllowList dictionary
    :return: RuleAllowList
        The output will contain an object of RuleAllowList
    """
    rule_allow_list = None
    if allow_list_dictionary:
        description = allow_list_dictionary["description"] if "description" in allow_list_dictionary else None
        regexes = None
        paths = None
        commits = None
        stopwords = None

        if "regexes" in allow_list_dictionary:
            regexes = ""
            regexes_array = allow_list_dictionary["regexes"]
            for index, regex in enumerate(regexes_array):
                if index + 1 < len(regexes_array):
                    regexes += regex + TOML_CUSTOM_DELIMITER
                else:
                    regexes += regex

        if "paths" in allow_list_dictionary:
            paths = ""
            paths_array = allow_list_dictionary["paths"]
            for index, path in enumerate(paths_array):
                if index + 1 < len(paths_array):
                    paths += path + TOML_CUSTOM_DELIMITER
                else:
                    paths += path

        if "commits" in allow_list_dictionary:
            commits_array = allow_list_dictionary["commits"]
            commits = ",".join(commits_array)

        if "stopwords" in allow_list_dictionary:
            stopword_array = allow_list_dictionary["stopwords"]
            stopwords = ",".join(stopword_array)

        rule_allow_list = RuleAllowList(
            description=description,
            regexes=regexes,
            paths=paths,
            commits=commits,
            stop_words=stopwords,
        )
    return rule_allow_list


def create_toml_rule_file(parsed_toml_dictionary: dict):
    """
        Create a TOML file from a dictionary
    :param parsed_toml_dictionary:
        TOML dictionary
    :return: toml_file
        Returns toml file
    """
    doc = document()
    doc.add(comment("This is a gitleaks configuration file."))
    doc.add(comment("Rules and allowlists are defined within this file."))
    doc.add(comment("Rules instruct gitleaks on what should be considered a secret."))
    doc.add(comment("Allowlists instruct gitleaks on what is allowed, i.e. not a secret."))
    doc.add(nl())

    if "title" in parsed_toml_dictionary:
        doc.add("title", parsed_toml_dictionary["title"])
    doc.add(nl())
    if "version" in parsed_toml_dictionary:
        doc.add("version", parsed_toml_dictionary["version"])
    doc.add(nl())

    # Global allow list table
    global_allow_list_table = create_allow_list_toml_table(input_dictionary=parsed_toml_dictionary, key="allowlist")
    if global_allow_list_table:
        doc.add("allowlist", global_allow_list_table)
        doc.add(nl())

    # Rules table
    rule_array_table = create_rule_array_toml_table(rule_dictionary=parsed_toml_dictionary)
    doc.add("rules", rule_array_table)

    toml_string = tomlkit.dumps(doc)
    with open(TEMP_RULE_FILE, "w", encoding="utf-8") as toml_file:
        toml_file.write(toml_string)
        toml_file.close()
    return toml_file


def get_multiline_array_for_toml_file(
    input_dictionary: dict, key: str, string_type: str, delimiter: str
) -> tomlkit.array():
    """
        Create multiline toml array for the input dictionary value
    :param input_dictionary:
        Input dictionary
    :param key:
        key of Input dictionary
    :param string_type:
        Multi Line Literal or Single Line Basic
    :param delimiter:
        TOML_CUSTOM_DELIMITER or ","
    :return: tomlkit.array
        The output will return a toml array
    """
    multiline_array = tomlkit.array()
    array_from_db = input_dictionary[key].split(delimiter)
    for value_str in array_from_db:
        multiline_array.append(String.from_raw(String.from_raw(value_str), string_type))
        multiline_array.multiline(True)
    return multiline_array


def create_allow_list_toml_table(input_dictionary: dict, key: str) -> table():
    """
        Create a TOML table for rule allow list
    :param input_dictionary:
        AllowList dictionary
     :param key:
        Key in AllowList dictionary
    :return: table
        Returns allow list TOML table
    """
    allow_list_table = None
    if key in input_dictionary:
        allow_list_table = table()
        allow_list_dict = input_dictionary[key]
        if "description" in allow_list_dict:
            allow_list_table.add("description", allow_list_dict["description"])
        if "paths" in allow_list_dict:
            multiline_path_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="paths",
                string_type=StringType.MLL,
                delimiter=TOML_CUSTOM_DELIMITER,
            )
            allow_list_table.add("paths", multiline_path_array)
        if "regexes" in allow_list_dict:
            multiline_regex_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="regexes",
                string_type=StringType.MLL,
                delimiter=TOML_CUSTOM_DELIMITER,
            )
            allow_list_table.add("regexes", multiline_regex_array)
        if "commits" in allow_list_dict:
            multiline_commit_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="commits",
                string_type=StringType.SLB,
                delimiter=",",
            )
            allow_list_table.add("commits", multiline_commit_array)
        if "stop_words" in allow_list_dict:
            multiline_stopword_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="stop_words",
                string_type=StringType.SLB,
                delimiter=",",
            )
            allow_list_table.add("stopwords", multiline_stopword_array)
    return allow_list_table


def create_rule_array_toml_table(rule_dictionary: dict) -> aot():
    """
       Create an array of table for rule list
    :param rule_dictionary:
        Rule dictionary
    :return: table
        Return an array of table
    """

    # Rule Table
    rule_array_table = aot()
    if "rules" in rule_dictionary:
        for rule_dict in rule_dictionary["rules"]:
            rule_table = table()
            if "id" in rule_dict:
                rule_table.add("id", rule_dict["id"])
            if "description" in rule_dict:
                rule_table.add("description", rule_dict["description"])
            if "entropy" in rule_dict:
                rule_table.add("entropy", rule_dict["entropy"])
            if "secret_group" in rule_dict:
                rule_table.add("secretGroup", rule_dict["secret_group"])
            if "regex" in rule_dict:
                rule_table.add("regex", String.from_raw(rule_dict["regex"], StringType.MLL))
            if "path" in rule_dict:
                rule_table.add("path", String.from_raw(rule_dict["path"], StringType.MLL))
            if "tags" in rule_dict:
                multiline_tag_array = get_multiline_array_for_toml_file(
                    input_dictionary=rule_dict,
                    key="tags",
                    string_type=StringType.SLB,
                    delimiter=",",
                )
                rule_table.add("tags", multiline_tag_array)
            if "keywords" in rule_dict:
                multiline_keyword_array = get_multiline_array_for_toml_file(
                    input_dictionary=rule_dict,
                    key="keywords",
                    string_type=StringType.SLB,
                    delimiter=",",
                )
                rule_table.add("keywords", multiline_keyword_array)

            # Rule Allow List Table
            if "allow_list" in rule_dict:
                allow_list_table = create_allow_list_toml_table(input_dictionary=rule_dict, key="allow_list")
                rule_table.append("allowlist", allow_list_table)

            rule_array_table.append(rule_table)
    return rule_array_table


def validate_uploaded_file_and_read_content(rule_file: File) -> str:
    """
       Validate the uploaded file and read content
    :param rule_file:
        File uploaded
    :return: content
        Return file content
    """
    file_name = os.path.splitext(rule_file.filename)[0]

    # File name validation
    is_valid_file_name = bool(re.match(FILE_NAME_REGEX, file_name))
    if not is_valid_file_name:
        raise HTTPException(500, detail=f"Invalid characters in File name - {file_name}")

    # File name max length validation
    if len(file_name) > 255:
        raise HTTPException(500, detail="File name value exceeds maximum length of 255 characters")

    # File extension validation
    if rule_file.content_type != "application/octet-stream" or not rule_file.filename.lower().endswith(
        ALLOWED_EXTENSION
    ):
        raise HTTPException(500, detail="Invalid document type, only TOML file is supported")

    # File size validation
    max_size: int = 1000000
    content = rule_file.file.read()
    file_size = len(content)
    if file_size > max_size:
        raise HTTPException(500, detail="File size exceeds the maximum limit 1 MB")

    toml_content = str(content, "utf-8")
    return toml_content



# Standard Library
from enum import Enum

# First Party
from resc_backend.constants import AZURE_DEVOPS, BITBUCKET, GITHUB_PUBLIC


class VCSProviders(str, Enum):
    AZURE_DEVOPS = AZURE_DEVOPS
    BITBUCKET = BITBUCKET
    GITHUB_PUBLIC = GITHUB_PUBLIC



PROJECT_QUEUE = "projects"
REPOSITORY_QUEUE = "repositories"

BITBUCKET = "BITBUCKET"
AZURE_DEVOPS = "AZURE_DEVOPS"
GITHUB_PUBLIC = "GITHUB_PUBLIC"

# RWS: RESC Web Service
RWS_VERSION_PREFIX = "/resc/v1"
RWS_ROUTE_REPOSITORIES = "/repositories"
RWS_ROUTE_SCANS = "/scans"
RWS_ROUTE_LAST_SCAN = "/last-scan"
RWS_ROUTE_FINDINGS = "/findings"
RWS_ROUTE_RULES = "/rules"
RWS_ROUTE_METRICS = "/metrics"
RWS_ROUTE_DETAILED_FINDINGS = "/detailed-findings"
RWS_ROUTE_TOTAL_COUNT_BY_RULE = "/total-count-by-rule"
RWS_ROUTE_BY_RULE = "/by-rule"
RWS_ROUTE_DETECTED_RULES = "/detected-rules"
RWS_ROUTE_FINDINGS_METADATA = "/findings-metadata"
RWS_ROUTE_FINDING_STATUS_COUNT = "/finding-status-count"
RWS_ROUTE_RULE_PACKS = "/rule-packs"
RWS_ROUTE_VCS = "/vcs-instances"


RWS_ROUTE_PERSONAL_AUDITS = "/personal-audits"
RWS_ROUTE_AUDIT_COUNT_BY_AUDITOR_OVER_TIME = "/audit-count-by-auditor-over-time"
RWS_ROUTE_AUDITED_COUNT_OVER_TIME = "/audited-count-over-time"
RWS_ROUTE_UN_TRIAGED_COUNT_OVER_TIME = "/un-triaged-count-over-time"
RWS_ROUTE_COUNT_BY_TIME = "/count-by-time"
RWS_ROUTE_COUNT_PER_VCS_PROVIDER_BY_WEEK = "/count-per-vcs-provider-by-week"
RWS_ROUTE_SUPPORTED_VCS_PROVIDERS = "/supported-vcs-providers"
RWS_ROUTE_SUPPORTED_STATUSES = "/supported-statuses"
RWS_ROUTE_DISTINCT_PROJECTS = "/distinct-projects"
RWS_ROUTE_DISTINCT_REPOSITORIES = "/distinct-repositories"
COMMON_TAG = "resc-common"

RWS_ROUTE_AUDIT = "/audit"

RWS_ROUTE_AUTH_CHECK = "/auth-check"

RWS_ROUTE_HEALTH = "/health"

REPOSITORIES_TAG = "resc-repositories"
SCANS_TAG = "resc-scans"
FINDINGS_TAG = "resc-findings"
RULES_TAG = "resc-rules"
RULE_PACKS_TAG = "resc-rule-packs"
HEALTH_TAG = "health"
VCS_TAG = "resc-vcs-instances"
METRICS_TAG = "resc-metrics"

DEFAULT_RECORDS_PER_PAGE_LIMIT = 100
MAX_RECORDS_PER_PAGE_LIMIT = 500

BASE_SCAN = "BASE"
INCREMENTAL_SCAN = "INCREMENTAL"

# Cache
CACHE_PREFIX = "resc-cache"
CACHE_NAMESPACE_FINDING = "namespace-finding"
CACHE_NAMESPACE_REPOSITORY = "namespace-repository"
CACHE_NAMESPACE_VCS_INSTANCE = "namespace-vcs-instance"
CACHE_NAMESPACE_RULE_PACK = "namespace-rule-pack"
CACHE_NAMESPACE_RULE = "namespace-rule"
CACHE_NAMESPACE_FINDING_STATUS = "namespace-finding-status"

TOML_CUSTOM_DELIMITER = "#custom-delimiter#"
TEMP_RULE_FILE = "/tmp/temp_resc_rule.toml"
ALLOWED_EXTENSION = ".toml"

# Logging for dummy-data-generator
LOG_FILE_DUMMY_DATA_GENERATOR = "dummy-data-gen.log"

# Logging
LOG_FILE_PATH_RABBITMQ = "rabbitmq_initialization.log"
LOG_FILE_CACHING = "redis_cache.log"
LOGGING_FILE = "logging.ini"

# Error message
ERROR_MESSAGE_500 = "Internal server error. Contact your system administrator"
ERROR_MESSAGE_503 = "Unable to communicate with DataBase, Please contact your system administrator"

# Redis Cache
REDIS_CACHE_EXPIRE = 60 * 60 * 24  # set to 24 hours

# HTTP Security Response Headers
STRICT_TRANSPORT_SECURITY = "max-age=31536000; includeSubDomains; preload"
CACHE_CONTROL = "no-cache, no-store"
CROSS_ORIGIN_RESOURCE_POLICY = "same-site"
REFERRER_POLICY = "same-origin"
X_PERMITTED_CROSS_DOMAIN_POLICIES = "none"
X_CONTENT_TYPE_OPTIONS = "nosniff"
X_FRAME_OPTIONS = "DENY"
X_XSS_PROTECTION = "1; mode=block"
CONTENT_SECURITY_POLICY = (
    "default-src 'none'; script-src 'self' https://cdn.jsdelivr.net 'unsafe-inline'; "
    "connect-src 'self'; img-src 'self' https://fastapi.tiangolo.com data:;style-src "
    "'self' https://fonts.googleapis.com https://cdn.jsdelivr.net 'unsafe-inline';"
    "frame-ancestors 'self'; form-action 'self';"
)



# Standard Library
from enum import Enum

# First Party
from resc_backend.constants import BASE_SCAN, INCREMENTAL_SCAN


class ScanType(str, Enum):
    BASE = BASE_SCAN
    INCREMENTAL = INCREMENTAL_SCAN



# Standard Library
from datetime import datetime

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, Request, status
from fastapi_cache.decorator import cache

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    CACHE_NAMESPACE_FINDING_STATUS,
    CACHE_NAMESPACE_RULE,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    FINDINGS_TAG,
    REDIS_CACHE_EXPIRE,
    RWS_ROUTE_AUDIT,
    RWS_ROUTE_BY_RULE,
    RWS_ROUTE_COUNT_BY_TIME,
    RWS_ROUTE_FINDINGS,
    RWS_ROUTE_SUPPORTED_STATUSES,
    RWS_ROUTE_TOTAL_COUNT_BY_RULE,
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import audit as audit_crud
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.helpers.resc_swagger_models import Model400, Model404
from resc_backend.resc_web_service.schema import audit as audit_schema
from resc_backend.resc_web_service.schema import finding as finding_schema
from resc_backend.resc_web_service.schema.date_count_model import DateCountModel
from resc_backend.resc_web_service.schema.date_filter import DateFilter
from resc_backend.resc_web_service.schema.finding import FindingRead
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel

router = APIRouter(prefix=f"{RWS_ROUTE_FINDINGS}", tags=[FINDINGS_TAG])


@router.get(
    "",
    response_model=PaginationModel[finding_schema.FindingRead],
    summary="Get findings",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the findings"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_all_findings(
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[finding_schema.FindingRead]:
    """
        Retrieve all findings objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [FindingRead]
        The output will contain a PaginationModel containing the list of FindingRead type objects,
        or an empty list if no finding was found
    """
    findings = finding_crud.get_findings(db_connection, skip=skip, limit=limit)

    total_findings = finding_crud.get_total_findings_count(db_connection)

    return PaginationModel[finding_schema.FindingRead](data=findings, total=total_findings, limit=limit, skip=skip)


@router.post(
    "",
    response_model=int,
    summary="Create a finding",
    status_code=status.HTTP_201_CREATED,
    responses={
        201: {"description": "Create new findings"},
        400: {"model": Model400, "description": "Error creating findings"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def create_findings(
    findings: list[finding_schema.FindingCreate],
    db_connection: Session = Depends(get_db_connection),
) -> int:
    """
          Create new findings

    - **db_connection**: Session of the database connection
    - **file_path**: file path
    - **line_number**: Line number
    - **commit_id**: commit hash
    - **commit_message**: Commit message
    - **commit_timestamp**: Commit timestamp
    - **author**: Author name
    - **email**: Email of the author
    - **event_sent_on**: event sent timestamp
    - **rule_name**: rule name
    - **repository_id**: repository id of the finding
    - **return**: int
          The output will contain the number of successful created findings
    """
    try:
        created_findings = finding_crud.create_findings(db_connection=db_connection, findings=findings)

        # Clear cache related to findings
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_FINDING)
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_RULE)
    except KeyError as err:
        raise HTTPException(status_code=400, detail=str(err)) from err
    return len(created_findings)


@router.get(
    "/{finding_id}",
    response_model=finding_schema.FindingRead,
    summary="Fetch a finding by ID",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve finding <finding_id>"},
        404: {"model": Model404, "description": "Finding <finding_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def read_finding(finding_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Read a finding by ID

    - **db_connection**: Session of the database connection
    - **finding_id**: ID of the finding for which details need to be fetched
    """
    db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    db_scan_findings = scan_finding_crud.get_scan_findings(db_connection, finding_id=finding_id)
    scan_ids = [x.scan_id for x in db_scan_findings]
    return FindingRead.create_from_db_entities(db_finding=db_finding, scan_ids=scan_ids)


@router.patch(
    "/{finding_id}",
    response_model=finding_schema.FindingRead,
    summary="Partially update a finding by ID",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Modify finding <finding_id>"},
        404: {"model": Model404, "description": "Finding <finding_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def patch_finding(
    finding_id: int,
    finding_update: finding_schema.FindingPatch,
    db_connection: Session = Depends(get_db_connection),
):
    """
        Partially update a finding by ID

    - **db_connection**: Session of the database connection
    - **finding_id**: ID of the finding for which details need to be updated
    - **event_sent_on**: Event sent timestamp
    """
    db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
    db_sca_findings = scan_finding_crud.get_scan_findings(db_connection, finding_id=finding_id)
    scan_ids = [x.scan_id for x in db_sca_findings]

    # Clear cache related to findings
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_FINDING)
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    return FindingRead.create_from_db_entities(
        finding_crud.patch_finding(db_connection, finding_id=finding_id, finding_update=finding_update),
        scan_ids,
    )


@router.delete(
    "/{finding_id}",
    summary="Delete a finding",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Delete finding <finding_id>"},
        404: {"model": Model404, "description": "Finding <finding_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def delete_finding(finding_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Delete a finding object

    - **db_connection**: Session of the database connection
    - **finding_id**: ID of the finding to delete
    - **return**: The output will contain a success or error message based on the success of the deletion
    """
    db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    finding_crud.delete_finding(db_connection, finding_id=finding_id, delete_related=True)

    # Clear cache related to findings
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_FINDING)
    return {"ok": True}


@router.get(
    f"{RWS_ROUTE_TOTAL_COUNT_BY_RULE}" "/{rule_name}",
    summary="Get total findings count by rule",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve total findings count of rule <rule_name>"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_total_findings_count_by_rule(rule_name: str, db_connection: Session = Depends(get_db_connection)):
    """
        Retrieve total findings count for a given rule

    - **db_connection**: Session of the database connection
    - **rule_name**: name of the rule
    """
    findings_filter = FindingsFilter(rule_names=[rule_name])
    return finding_crud.get_total_findings_count(db_connection, findings_filter=findings_filter)


@router.get(
    f"{RWS_ROUTE_BY_RULE}" "/{rule_name}",
    response_model=PaginationModel[finding_schema.FindingRead],
    summary="Get findings by rule",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the findings of rule <rule_name>"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_findings_by_rule(
    rule_name: str,
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[finding_schema.FindingRead]:
    """
        Retrieve all findings objects paginated by rule

    - **db_connection**: Session of the database connection
    - **rule_name**: Name of the rule to filter the findings by
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [FindingRead]
        The output will contain a PaginationModel containing the list of FindingRead type objects,
        or an empty list if no finding was found for the given rule
    """
    findings = finding_crud.get_findings_by_rule(db_connection, skip=skip, limit=limit, rule_name=rule_name)
    total_findings = finding_crud.get_total_findings_count(
        db_connection, findings_filter=FindingsFilter(rule_names=[rule_name])
    )
    return PaginationModel[finding_schema.FindingRead](data=findings, total=total_findings, limit=limit, skip=skip)


@router.post(
    f"{RWS_ROUTE_AUDIT}/",
    response_model=int,
    summary="audit single/multiple findings",
    status_code=status.HTTP_201_CREATED,
    responses={
        201: {"description": "Audit(s) successfully saved"},
        404: {"model": Model404, "description": "Finding <finding_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def audit_findings(
    request: Request,
    audit: audit_schema.AuditMultiple,
    db_connection: Session = Depends(get_db_connection),
) -> int:
    """
        Audit single/multiple findings, updating the status and comment

    - **db_connection**: Session of the database connection
    - **finding_ids**: List of finding IDs for which audit to be performed
    - **status**: Status of the finding, Valid values are NOT_ANALYZED, NOT_ACCESSIBLE,
                  CLARIFICATION_REQUIRED, FALSE_POSITIVE, TRUE_POSITIVE
    - **comment**: Comment
    - **return**: int
        The output will contain count of successful saved audits
    """
    audits = []
    for finding_id in audit.finding_ids:
        db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
        if db_finding is None:
            raise HTTPException(status_code=404, detail=f"Finding {finding_id} not found")
        audits.append(
            audit_crud.create_audit(
                db_connection=db_connection,
                finding_id=db_finding.id_,
                auditor=request.user,
                status=audit.status,
                comment=audit.comment,
            )
        )

        # Clear cache related to findings
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_FINDING)
    return len(audits)


@router.get(
    "/{finding_id}" f"{RWS_ROUTE_AUDIT}",
    response_model=PaginationModel[audit_schema.AuditRead],
    summary="Get audit(s) for finding",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the audit entries for a finding"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_finding_audits(
    finding_id: int,
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[audit_schema.AuditRead]:
    """
        Retrieve all audit objects paginated for a finding

    - **db_connection**: Session of the database connection
    - **finding_id**: id of the finding to get the audit for
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [AuditRead]
        The output will contain a PaginationModel containing the list of AuditRead type objects,
        or an empty list if no audit info was found
    """
    audits = audit_crud.get_finding_audits(db_connection, skip=skip, limit=limit, finding_id=finding_id)
    total_audits = audit_crud.get_finding_audits_count(db_connection, finding_id=finding_id)
    return PaginationModel[audit_schema.AuditRead](data=audits, total=total_audits, limit=limit, skip=skip)


@router.get(
    f"{RWS_ROUTE_SUPPORTED_STATUSES}/",
    response_model=list[str],
    summary="Get all supported statuses for findings",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the supported statuses for the findings"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING_STATUS, expire=REDIS_CACHE_EXPIRE)
def get_supported_statuses() -> list[str]:
    """
        Retrieve all supported statuses for findings

    - **return**: List[str]
        The output will contain a list of strings of unique statuses supported
    """
    supported_finding_statuses = [finding_status for finding_status in FindingStatus if finding_status]
    return supported_finding_statuses


@router.get(
    f"{RWS_ROUTE_COUNT_BY_TIME}/" "{time_type}",
    response_model=PaginationModel[DateCountModel],
    summary="Get all the findings by time period",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the findings by time-period <time_type>"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_count_by_time(
    time_type: DateFilter,
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    start_date_time: datetime | None = Query(None),
    end_date_time: datetime | None = Query(None),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[DateCountModel]:
    """
        Retrieve all findings count by time period objects paginated

    - **db_connection**: Session of the database connection
    - **time_type**: required, filter on time type. Available values: month, week, day
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **start_date_time**: Optional, filter on start date
    - **end_date_time**: Optional, filter on end date
    - **return**: PaginationModel[DateCountModel]
        The output will contain a PaginationModel containing the list of DateCountModel type objects,
        or an empty list if no data was found
    """
    date_counts = []
    findings = finding_crud.get_findings_count_by_time(
        db_connection=db_connection,
        date_type=time_type,
        start_date_time=start_date_time,
        end_date_time=end_date_time,
        skip=skip,
        limit=limit,
    )
    total_findings = finding_crud.get_findings_count_by_time_total(
        db_connection=db_connection,
        date_type=time_type,
        start_date_time=start_date_time,
        end_date_time=end_date_time,
    )
    for finding in findings:
        if time_type == DateFilter.MONTH:
            date_count = DateCountModel(finding_count=finding[2], date_lable=f"{finding[0]}-{finding[1]}")
        elif time_type == DateFilter.WEEK:
            date_count = DateCountModel(finding_count=finding[2], date_lable=f"{finding[0]}-W{finding[1]}")
        elif time_type == DateFilter.DAY:
            date_count = DateCountModel(
                finding_count=finding[3],
                date_lable=f"{finding[0]}-{finding[1]}-{finding[2]}",
            )

        date_counts.append(date_count)

    return PaginationModel[DateCountModel](data=date_counts, total=total_findings, limit=limit, skip=skip)



# Standard Library
import datetime

# Third Party
from pydantic import BaseModel, conint, conlist, constr

# First Party
from resc_backend.resc_web_service.schema.finding_status import FindingStatus


class AuditMultiple(BaseModel):
    finding_ids: conlist(conint(gt=0), min_items=1, max_items=500)
    status: FindingStatus
    comment: constr(max_length=255)


class AuditRead(BaseModel):
    id_: conint(gt=0)
    status: FindingStatus
    auditor: constr(max_length=250)
    comment: constr(max_length=255) | None = None
    timestamp: datetime.datetime

    class Config:
        orm_mode = True



# Standard Library
from datetime import datetime, UTC

# Third Party
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    String,
    text,
)

# First Party
from resc_backend.constants import BASE_SCAN
from resc_backend.db.model import Base
from resc_backend.db.model.rule_pack import DBrulePack
from resc_backend.resc_web_service.schema.scan_type import ScanType

REPOSITORY_ID = "repository.id"


class DBscan(Base):
    __tablename__ = "scan"
    id_ = Column("id", Integer, primary_key=True)
    repository_id = Column(Integer, ForeignKey(REPOSITORY_ID))
    rule_pack = Column(String(100), ForeignKey(DBrulePack.version), nullable=False)
    scan_type = Column(Enum(ScanType), default=ScanType.BASE, server_default=BASE_SCAN, nullable=False)
    last_scanned_commit = Column(String(100), nullable=False)
    timestamp = Column(DateTime, nullable=False, default=datetime.now(UTC))
    increment_number = Column(Integer, server_default=text("0"), default=0, nullable=False)
    is_latest = Column(Boolean, nullable=False, default=False, server_default=text("0"))

    def __init__(
        self,
        repository_id: int,
        scan_type: ScanType,
        last_scanned_commit: str,
        timestamp: datetime,
        increment_number: int,
        rule_pack: str,
        is_latest: bool,
    ):
        self.repository_id = repository_id
        self.scan_type = scan_type
        self.last_scanned_commit = last_scanned_commit
        self.timestamp = timestamp
        self.increment_number = increment_number
        self.rule_pack = rule_pack
        self.is_latest = is_latest

    @staticmethod
    def create_from_metadata(
        timestamp: datetime,
        scan_type: ScanType,
        last_scanned_commit: str,
        increment_number: int,
        rule_pack: str,
        repository_id: int,
        is_latest: bool,
    ):
        db_scan = DBscan(
            timestamp=timestamp,
            scan_type=scan_type,
            last_scanned_commit=last_scanned_commit,
            increment_number=increment_number,
            rule_pack=rule_pack,
            repository_id=repository_id,
            is_latest=is_latest,
        )
        return db_scan



# Standard Library

# Third Party
from pydantic import BaseModel, conint, constr


class RuleAllowListBase(BaseModel):
    description: constr(max_length=2000) | None
    regexes: str | None
    paths: str | None
    commits: str | None
    stop_words: str | None


class RuleAllowListCreate(RuleAllowListBase):
    @classmethod
    def create_from_base_class(cls, base_object: RuleAllowListBase):
        return cls(**(dict(base_object)))


class RuleAllowList(RuleAllowListBase):
    pass


class RuleAllowListRead(RuleAllowListBase):
    id_: conint(gt=0)

    class Config:
        orm_mode = True



# Standard Library
import logging

# Third Party
from fastapi import APIRouter, status

# First Party
from resc_backend.constants import (
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    HEALTH_TAG,
    RWS_ROUTE_HEALTH,
)

router = APIRouter(tags=[HEALTH_TAG])

logger = logging.getLogger(__name__)


@router.get(
    f"{RWS_ROUTE_HEALTH}",
    summary="Health check",
    description="Retrieve the health status of RESC APIs",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve the health status"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def health_check():
    return {"status": "OK"}



# Standard Library
import logging

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, status
from sqlalchemy.exc import IntegrityError

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    CACHE_NAMESPACE_RULE,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    RWS_ROUTE_DETECTED_RULES,
    RWS_ROUTE_FINDINGS,
    RWS_ROUTE_SCANS,
    SCANS_TAG,
)
from resc_backend.db.connection import Session
from resc_backend.db.model import DBscanFinding
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.helpers.resc_swagger_models import Model400, Model404
from resc_backend.resc_web_service.schema import finding as finding_schema
from resc_backend.resc_web_service.schema import scan as scan_schema
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel
from resc_backend.resc_web_service.schema.scan_type import ScanType

router = APIRouter(prefix=f"{RWS_ROUTE_SCANS}", tags=[SCANS_TAG])
logger = logging.getLogger(__name__)


@router.get(
    "",
    response_model=PaginationModel[scan_schema.ScanRead],
    summary="Get scans",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the scan objects"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_all_scans(
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[scan_schema.ScanRead]:
    """
        Retrieve all scan objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [ScanRead]
        The output will contain a PaginationModel containing the list of ScanRead type objects,
        or an empty list if no scan was found
    """
    scans = scan_crud.get_scans(db_connection, skip=skip, limit=limit)

    total_scans = scan_crud.get_scans_count(db_connection)

    return PaginationModel[scan_schema.ScanRead](data=scans, total=total_scans, limit=limit, skip=skip)


@router.post(
    "",
    response_model=scan_schema.ScanRead,
    summary="Create a scan",
    status_code=status.HTTP_201_CREATED,
    responses={
        201: {"description": "Create a new scan"},
        400: {"model": Model400, "description": "Error creating a new scan"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def create_scan(scan: scan_schema.ScanCreate, db_connection: Session = Depends(get_db_connection)):
    """
        Create a scan with all the information

    - **db_connection**: Session of the database connection
    - **scan_type**: scan type, supported values are BASE or INCREMENTAL
    - **last_scanned_commit**: last scanned commit hash
    - **timestamp**: creation timestamp
    - **increment_number**: scan increment number
    - **rule_pack**: rule pack version
    - **repository_id**: repository id
    """
    # Determine the increment number if needed and not supplied
    if scan.scan_type == ScanType.INCREMENTAL and (not scan.increment_number or scan.increment_number <= 0):
        last_scan = scan_crud.get_latest_scan_for_repository(db_connection, repository_id=scan.repository_id)
        new_increment = last_scan.increment_number + 1
        scan.increment_number = new_increment

    try:
        created_scan = scan_crud.create_scan(db_connection=db_connection, scan=scan)
    except IntegrityError as err:
        logger.debug(f"Error creating new scan object: {err}")
        raise HTTPException(status_code=400, detail="Error creating new scan object") from err
    return created_scan


@router.get(
    "/{scan_id}",
    response_model=scan_schema.ScanRead,
    summary="Fetch a scan by ID",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve scan <scan_id>"},
        404: {"model": Model404, "description": "Scan <scan_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def read_scan(scan_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Read a scan by ID

    - **db_connection**: Session of the database connection
    - **scan_id**: ID of the scan for which details need to be fetched
    """
    db_scan = scan_crud.get_scan(db_connection, scan_id=scan_id)
    if db_scan is None:
        raise HTTPException(status_code=404, detail="Scan not found")
    return db_scan


@router.put(
    "/{scan_id}",
    response_model=scan_schema.ScanRead,
    summary="Update an existing scan",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Update scan <scan_id>"},
        404: {"model": Model404, "description": "Scan <scan_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def update_scan(
    scan_id: int,
    scan: scan_schema.ScanCreate,
    db_connection: Session = Depends(get_db_connection),
):
    """
        Update an existing scan

    - **db_connection**: Session of the database connection
    - **scan_type**: scan type, supported values are BASE or INCREMENTAL
    - **last_scanned_commit**: last scanned commit
    - **timestamp**: scan timestamp
    - **increment_number**: scan increment number
    - **rule_pack**: rule pack version
    - **repository_id**: repository id

    """
    db_scan = scan_crud.get_scan(db_connection, scan_id=scan_id)
    if db_scan is None:
        raise HTTPException(status_code=404, detail="Scan not found")
    return scan_crud.update_scan(db_connection=db_connection, scan_id=scan_id, scan=scan)


@router.delete(
    "/{scan_id}",
    summary="Delete a scan",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Delete scan <scan_id>"},
        404: {"model": Model404, "description": "Scan <scan_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def delete_scan(scan_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Delete a scan object

    - **db_connection**: Session of the database connection
    - **scan_id**: ID of the scan to delete
    - **return**: The output will contain a success or error message based on the success of the deletion
    """
    db_scan = scan_crud.get_scan(db_connection, scan_id=scan_id)
    if db_scan is None:
        raise HTTPException(status_code=404, detail="Scan not found")
    scan_crud.delete_scan(
        db_connection,
        repository_id=db_scan.repository_id,
        scan_id=scan_id,
        delete_related=True,
    )
    return {"ok": True}


@router.post(
    "/{scan_id}" f"{RWS_ROUTE_FINDINGS}",
    response_model=int,
    summary="Create scan findings",
    status_code=status.HTTP_201_CREATED,
    responses={
        201: {"description": "Create findings and their associated scan_findings for scan <scan_id>"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def create_scan_findings(
    scan_id: int,
    findings: list[finding_schema.FindingCreate],
    db_connection: Session = Depends(get_db_connection),
) -> int:
    """
        Creates findings and their associated scan_findings for a given scan

    - **db_connection**: Session of the database connection
    - **scan_id**:  Id of the scan for which findings need to be inserted
    - **file_path**: file path
    - **line_number**: Line number
    - **column_start**: Column start
    - **column_end**: Column end
    - **commit_id**: commit hash
    - **commit_message**: Commit message
    - **commit_timestamp**: Commit timestamp
    - **author**: Author name
    - **email**: Email of the author
    - **status**: Status of the finding, Valid values are NOT_ANALYZED, NOT_ACCESSIBLE,
                  CLARIFICATION_REQUIRED, FALSE_POSITIVE, TRUE_POSITIVE
    - **comment**: Comment
    - **event_sent_on**: event sent timestamp
    - **rule_name**: rule name
    - **repository_id**: repository id of the finding
    - **return**: [FindingRead]
        The output will contain a PaginationModel containing the list of FindingRead type objects,
        or an empty list if no scan was found
    """

    created_findings = finding_crud.create_findings(db_connection=db_connection, findings=findings)
    scan_findings = []
    for finding in created_findings:
        scan_finding = DBscanFinding(finding_id=finding.id_, scan_id=scan_id)
        scan_findings.append(scan_finding)

    _ = scan_finding_crud.create_scan_findings(db_connection=db_connection, scan_findings=scan_findings)

    # Clear cache related to findings
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_FINDING)
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_RULE)

    return len(created_findings)


@router.get(
    "/{scan_id}" f"{RWS_ROUTE_FINDINGS}",
    response_model=PaginationModel[finding_schema.FindingRead],
    summary="Get scan findings associated with a scan ID",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve findings associated with scan <scan_id>"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_scan_findings(
    scan_id: int,
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    rules: list[str] = Query([], pattern=r"^[A-z0-9 .\-_%]*$", alias="rule", title="rule"),
    statuses: list[FindingStatus] = Query(None, alias="status", title="status"),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[finding_schema.FindingRead]:
    """
        Retrieve all finding objects paginated related to a scan_id

    - **db_connection**: Session of the database connection
    - **scan_id**: Id of the scan for which to retrieve the findings
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **rules**: optional, filter on rule name. It is used as a string contains filter
    - **statuses**:  optional, filter on status of findings
    - **return**: [FindingRead]
        The output will contain a PaginationModel containing the list of FindingRead type objects,
        or an empty list if no scan was found
    """
    findings = finding_crud.get_scans_findings(
        db_connection,
        scan_ids=[scan_id],
        skip=skip,
        limit=limit,
        rules_filter=rules,
        statuses_filter=statuses,
    )

    findings_filter = FindingsFilter(scan_ids=[scan_id], rule_names=rules, finding_statuses=statuses)
    total_findings = finding_crud.get_total_findings_count(db_connection, findings_filter=findings_filter)

    return PaginationModel[finding_schema.FindingRead](data=findings, total=total_findings, limit=limit, skip=skip)


@router.get(
    f"{RWS_ROUTE_FINDINGS}/",
    response_model=PaginationModel[finding_schema.FindingRead],
    summary="Get scan findings",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve findings associated with scan <scan_id>"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_scans_findings(
    scan_ids: list[int] = Query([], alias="scan_id", title="Scan ids"),
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    rules: list[str] = Query([], pattern=r"^[A-z0-9 .\-_%]*$", alias="rule", title="rule"),
    statuses: list[FindingStatus] = Query(None, alias="status", title="status"),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[finding_schema.FindingRead]:
    """
        Retrieve all finding objects paginated related to a scan_id

    - **db_connection**: Session of the database connection
    - **scan_ids**: Optional, List of scan IDs for which findings to be retrieved
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **rule**: optional, filter on rule name. It is used as a string contains filter
    - **statuses**:  optional, filter on status of findings
    - **return**: [FindingRead]
        The output will contain a PaginationModel containing the list of FindingRead type objects,
        or an empty list if no scan was found
    """
    findings = finding_crud.get_scans_findings(
        db_connection,
        scan_ids=scan_ids,
        skip=skip,
        limit=limit,
        rules_filter=rules,
        statuses_filter=statuses,
    )

    findings_filter = FindingsFilter(scan_ids=scan_ids, rule_names=rules, finding_statuses=statuses)
    total_findings = finding_crud.get_total_findings_count(db_connection, findings_filter=findings_filter)

    return PaginationModel[finding_schema.FindingRead](data=findings, total=total_findings, limit=limit, skip=skip)


@router.get(
    f"{RWS_ROUTE_DETECTED_RULES}/",
    response_model=list[str],
    summary="Get unique rules from scans",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the unique rules associated with specified scans"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_distinct_rules_from_scans(
    scan_ids: list[int] = Query([], alias="scan_id", title="Scan ids"),
    db_connection: Session = Depends(get_db_connection),
) -> list[str]:
    """
        Retrieve all uniquely detected rules for given scans

    - **db_connection**: Session of the database connection
    - **scan_ids**: scan ids for which to retrieve the unique rules
    - **return**: List[str]
        The output will contain a list of strings of unique rules for given scan ids
    """
    return_rules = []
    if scan_ids:
        rules = finding_crud.get_distinct_rules_from_scans(db_connection, scan_ids=scan_ids)
        for rule in rules:
            return_rules.append(rule.rule_name)
    return return_rules



# Third Party
from pydantic import BaseModel


class VcsProviderFindingCount(BaseModel):
    AZURE_DEVOPS: int = 0
    BITBUCKET: int = 0
    GITHUB_PUBLIC: int = 0


class FindingCountOverTime(BaseModel):
    time_period: str
    vcs_provider_finding_count: VcsProviderFindingCount = VcsProviderFindingCount()
    total: int = 0



# Standard Library
import datetime

# Third Party
from pydantic import BaseModel, conint, constr

RULE_PACK_VERSION_REGEX = r"^\d+(?:\.\d+){2}$"


class RulePackBase(BaseModel):
    version: constr(regex=RULE_PACK_VERSION_REGEX)
    active: bool = False
    global_allow_list: conint(gt=0) | None


class RulePackCreate(RulePackBase):
    version: constr(regex=RULE_PACK_VERSION_REGEX)
    global_allow_list: conint(gt=0) | None

    @classmethod
    def create_from_base_class(cls, base_object: RulePackBase, global_allow_list: int):
        return cls(**(dict(base_object)), global_allow_list=global_allow_list)


class RulePack(RulePackBase):
    pass


class RulePackRead(RulePackBase):
    version: constr(regex=RULE_PACK_VERSION_REGEX)
    created: datetime.datetime

    class Config:
        orm_mode = True



# Third Party
from sqlalchemy import Column, ForeignKey, Integer

# First Party
from resc_backend.db.model import Base


class DBruleTag(Base):
    __tablename__ = "rule_tag"
    rule_id = Column(Integer, ForeignKey("rules.id"), primary_key=True)
    tag_id = Column(Integer, ForeignKey("tag.id"), primary_key=True)

    def __init__(self, rule_id: int, tag_id: int):
        self.rule_id = rule_id
        self.tag_id = tag_id



# Standard Library
import logging
from datetime import datetime, timedelta, UTC

# Third Party
from fastapi import APIRouter, Depends, Query, Request, status
from fastapi_cache.decorator import cache
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    METRICS_TAG,
    REDIS_CACHE_EXPIRE,
    RWS_ROUTE_AUDIT_COUNT_BY_AUDITOR_OVER_TIME,
    RWS_ROUTE_AUDITED_COUNT_OVER_TIME,
    RWS_ROUTE_COUNT_PER_VCS_PROVIDER_BY_WEEK,
    RWS_ROUTE_METRICS,
    RWS_ROUTE_PERSONAL_AUDITS,
    RWS_ROUTE_UN_TRIAGED_COUNT_OVER_TIME,
)
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import audit as audit_crud
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.schema.audit_count_over_time import (
    AuditCountOverTime,
)
from resc_backend.resc_web_service.schema.finding_count_over_time import (
    FindingCountOverTime,
)
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.personal_audit_metrics import (
    PersonalAuditMetrics,
)
from resc_backend.resc_web_service.schema.time_period import TimePeriod
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(prefix=f"{RWS_ROUTE_METRICS}", tags=[METRICS_TAG])
logger = logging.getLogger(__name__)


@router.get(
    f"{RWS_ROUTE_AUDITED_COUNT_OVER_TIME}",
    response_model=list[FindingCountOverTime],
    summary="Get count of audit status over time for given weeks per vcs provider",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve count of audit status over time for given weeks per vcs provider"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_finding_audit_count_over_time(
    db_connection: Session = Depends(get_db_connection),
    weeks: int | None = Query(default=13, ge=1),
    audit_status: FindingStatus | None = Query(default=FindingStatus.TRUE_POSITIVE.value),
) -> list[FindingCountOverTime]:
    """
        Retrieve count of audited findings over time for given weeks per vcs provider
    - **db_connection**: Session of the database connection
    - **weeks**: Nr of weeks for which to retrieve the audit status count
    - **audit_status**: audit status for which to retrieve the counts, defaults to True positive
    - **return**: [DateCountModel]
        The output will contain a list of DateCountModel type objects
    """
    audit_counts = finding_crud.get_finding_audit_status_count_over_time(
        db_connection=db_connection, status=audit_status, weeks=weeks
    )
    output = convert_rows_to_finding_count_over_time(count_over_time=audit_counts, weeks=weeks)
    return output


@router.get(
    f"{RWS_ROUTE_COUNT_PER_VCS_PROVIDER_BY_WEEK}",
    response_model=list[FindingCountOverTime],
    summary="Get count of findings over time for given weeks per vcs provider",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve count of findings over time for given weeks per vcs provider"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_finding_total_count_over_time(
    db_connection: Session = Depends(get_db_connection),
    weeks: int | None = Query(default=13, ge=1),
) -> list[FindingCountOverTime]:
    """
        Retrieve count of findings over time for given weeks per vcs provider
    - **db_connection**: Session of the database connection
    - **weeks**: Nr of weeks for which to retrieve the audit status count
    - **audit_status**: audit status for which to retrieve the counts, defaults to True positive
    - **return**: [DateCountModel]
        The output will contain a list of DateCountModel type objects
    """
    audit_counts = finding_crud.get_finding_count_by_vcs_provider_over_time(db_connection=db_connection, weeks=weeks)
    output = convert_rows_to_finding_count_over_time(count_over_time=audit_counts, weeks=weeks)
    return output


@router.get(
    f"{RWS_ROUTE_UN_TRIAGED_COUNT_OVER_TIME}",
    response_model=list[FindingCountOverTime],
    summary="Get count of UnTriaged findings over time for given weeks per vcs provider",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve count of UnTriaged findings over time for given weeks per vcs provider"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_finding_un_triaged_count_over_time(
    db_connection: Session = Depends(get_db_connection),
    weeks: int | None = Query(default=13, ge=1),
) -> list[FindingCountOverTime]:
    """
        Retrieve count of UnTriaged findings over time for given weeks per vcs provider
    - **db_connection**: Session of the database connection
    - **weeks**: Nr of weeks for which to retrieve the audit status count
    - **audit_status**: audit status for which to retrieve the counts, defaults to True positive
    - **return**: [DateCountModel]
        The output will contain a list of DateCountModel type objects
    """
    audit_counts = finding_crud.get_un_triaged_finding_count_by_vcs_provider_over_time(
        db_connection=db_connection, weeks=weeks
    )
    output = convert_rows_to_finding_count_over_time(count_over_time=audit_counts, weeks=weeks)
    return output


def convert_rows_to_finding_count_over_time(count_over_time: dict, weeks: int) -> list[FindingCountOverTime]:
    """
        Convert the rows from the database to the format of list[FindingCountOverTime]
    :param count_over_time:
        rows from the database
    :param weeks:
        number fo weeks that are in the data
    :return: output
        list[FindingCountOverTime]
    """
    # Define the vcs provider types and finding statuses
    vcs_provider_types = list(VCSProviders)

    # create defaults with 0 value
    week_groups = {}
    for week in range(0, weeks):
        nth_week = datetime.now(UTC) - timedelta(weeks=week)
        week = f"{nth_week.isocalendar().year} W{nth_week.isocalendar().week:02d}"
        week_groups[week] = {vcs_provider_type: 0 for vcs_provider_type in vcs_provider_types + ["total"]}

    # loop over the counts from the database
    for data in count_over_time:
        week = f"{getattr(data, 'year')} W{getattr(data, 'week'):02d}"
        finding_count = getattr(data, "finding_count")

        week_groups[week][getattr(data, "provider_type")] += finding_count
        week_groups[week]["total"] += finding_count

    # Convert to the output format
    output = []
    for week in sorted(week_groups.keys()):
        week_data = FindingCountOverTime(time_period=week, total=week_groups[week]["total"])
        for vcs_provider_type in vcs_provider_types:
            setattr(
                week_data.vcs_provider_finding_count,
                vcs_provider_type,
                week_groups[week][vcs_provider_type],
            )

        output.append(week_data)
    return output


@router.get(
    f"{RWS_ROUTE_AUDIT_COUNT_BY_AUDITOR_OVER_TIME}",
    response_model=list[AuditCountOverTime],
    summary="Get count of Audits by Auditor over time for given weeks",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve count of Audits by Auditor over time for given weeks"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_audit_count_by_auditor_over_time(
    db_connection: Session = Depends(get_db_connection),
    weeks: int | None = Query(default=13, ge=1),
) -> list[AuditCountOverTime]:
    """
        Retrieve count of Audits by Auditor over time for given weeks
    - **db_connection**: Session of the database connection
    - **weeks**: Nr of weeks for which to retrieve the audit counts
    - **return**: [AuditCountOverTime]
        The output will contain a list of AuditCountOverTime type objects
    """
    audit_counts = audit_crud.get_audit_count_by_auditor_over_time(db_connection=db_connection, weeks=weeks)

    # get the unique auditors from the data
    auditors_default = {}
    for audit in audit_counts:
        auditors_default[getattr(audit, "auditor")] = 0

    # default to 0 per auditor for all weeks in range
    weekly_audit_counts = {}
    for week in range(0, weeks):
        nth_week = datetime.now(UTC) - timedelta(weeks=week)
        week = f"{nth_week.isocalendar().year} W{nth_week.isocalendar().week:02d}"
        weekly_audit_counts[week] = AuditCountOverTime(time_period=week, audit_by_auditor_count=dict(auditors_default))
    weekly_audit_counts = dict(sorted(weekly_audit_counts.items()))

    # set the counts based on the data from the database
    for audit in audit_counts:
        audit_week = f"{getattr(audit, 'year')} W{getattr(audit, 'week'):02d}"
        if audit_week in weekly_audit_counts:
            weekly_audit_counts.get(audit_week).audit_by_auditor_count[getattr(audit, "auditor")] = getattr(
                audit, "audit_count"
            )
            weekly_audit_counts.get(audit_week).total += getattr(audit, "audit_count")

    sorted_weekly_audit_counts = dict(sorted(weekly_audit_counts.items()))
    output = list(sorted_weekly_audit_counts.values())
    return output


@router.get(
    f"{RWS_ROUTE_PERSONAL_AUDITS}",
    response_model=PersonalAuditMetrics,
    summary="Get personal audit metrics",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Get personal audit metrics"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(
    namespace=CACHE_NAMESPACE_FINDING,
    expire=REDIS_CACHE_EXPIRE,
    key_builder=CacheManager.personalized_key_builder,
)
def get_personal_audit_metrics(
    request: Request, db_connection: Session = Depends(get_db_connection)
) -> PersonalAuditMetrics:
    """
        Retrieve personal audit metrics
    - **db_connection**: Session of the database connection
    - **return**: [DateCountModel]
        The output will contain a PersonalAuditMetrics type objects
    """
    audit_counts = PersonalAuditMetrics()
    audit_counts.today = audit_crud.get_personal_audit_count(
        db_connection=db_connection, auditor=request.user, time_period=TimePeriod.DAY
    )
    audit_counts.current_week = audit_crud.get_personal_audit_count(
        db_connection=db_connection, auditor=request.user, time_period=TimePeriod.WEEK
    )
    audit_counts.last_week = audit_crud.get_personal_audit_count(
        db_connection=db_connection,
        auditor=request.user,
        time_period=TimePeriod.LAST_WEEK,
    )
    audit_counts.current_month = audit_crud.get_personal_audit_count(
        db_connection=db_connection, auditor=request.user, time_period=TimePeriod.MONTH
    )
    audit_counts.current_year = audit_crud.get_personal_audit_count(
        db_connection=db_connection, auditor=request.user, time_period=TimePeriod.YEAR
    )
    audit_counts.forever = audit_crud.get_personal_audit_count(
        db_connection=db_connection,
        auditor=request.user,
        time_period=TimePeriod.FOREVER,
    )

    audit_counts.rank_current_week = determine_audit_rank_current_week(
        auditor=request.user, db_connection=db_connection
    )
    return audit_counts


def determine_audit_rank_current_week(auditor: str, db_connection: Session) -> int:
    """
        Retrieve personal audit ranking this week, compared to other auditors
    - **db_connection**: Session of the database connection
    - **auditor**: id of the auditor
    - **return**: int
        The output will be an integer nr of the ranking this week, defaulting to 0 if no audit was done by the auditor
    """
    audit_rank = 0
    audit_counts_db = audit_crud.get_audit_count_by_auditor_over_time(db_connection=db_connection, weeks=0)

    auditor_counts = {}
    for audit in audit_counts_db:
        auditor_counts[getattr(audit, "auditor")] = getattr(audit, "audit_count")

    sorted_auditor_counts = sorted(auditor_counts.items(), key=lambda x: x[1], reverse=True)
    for auditor_count in dict(sorted_auditor_counts):
        audit_rank += 1
        if auditor_count == auditor:
            return audit_rank
    return 0



# Standard Library
import os
from dataclasses import dataclass


@dataclass
class EnvironmentVariable:
    key_name: str
    help_text: str
    default: str = ""
    required: bool = False


def validate_environment(env_variables):
    missing = []
    values = {}
    for env_variable in env_variables:
        value = os.environ.get(env_variable.key_name, env_variable.default)
        if not value and env_variable.required:
            missing.append(f"{env_variable.key_name}: {env_variable.help_text}")
        else:
            values[env_variable.key_name] = value

    if missing:
        raise OSError(f"The following env variables need to be set: {', '.join(missing)}")
    return values



# Standard Library

# Third Party
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query

# First Party
from resc_backend.db.model import DBrule, DBruleTag, DBtag


def create_rule_tag(db_connection: Session, rule_id: int, tags: list[str]) -> list[DBruleTag]:
    """
    Create rule tag entries, linking / creating tag names to a rule
    :param db_connection:
        Session of the database connection
    :param rule_id:
        ID of the rule to link the tags to
    :param tags:
        List of string type containing tags to link to the rule
    :return: [DBruleTag]
        The output will contain a list of DBruleTag objects
    """
    db_tags = create_tags_if_not_exists(db_connection, tags)

    db_rule_tags = []
    for db_tag in db_tags:
        db_rule_tag = DBruleTag(rule_id=rule_id, tag_id=db_tag.id_)
        db_rule_tags.append(db_rule_tag)

    if db_rule_tags:
        db_connection.add_all(db_rule_tags)
        db_connection.flush()
        db_connection.commit()
    return db_rule_tags


def create_tags_if_not_exists(db_connection: Session, tags: list[str]) -> list[DBtag]:
    """
    Create tags if they don't exist or select existing
    :param db_connection:
        Session of the database connection
    :param tags:
        List of string type containing tags to create if they don't exist
    :return: [DBtag]
        The output will contain a list of tag objects
    """
    # Query the database to see if the tags objects exists
    db_tags = db_connection.query(DBtag).filter(DBtag.name.in_(tags)).all()
    if db_tags is not None:
        if len(db_tags) == len(tags):
            # all tags are in the db no need to create them
            return db_tags
        tags_known = [x.name for x in db_tags]
        tags_to_create = [tag for tag in tags if tag not in tags_known]
        # Created the tags not known
        created_tags = create_tags(db_connection, tags_to_create)
        db_tags.extend(created_tags)
    else:
        # None of the tags are known, create them all
        db_tags = create_tags(db_connection, tags)

    return db_tags


def create_tags(db_connection: Session, tags: list[str]) -> list[DBtag]:
    """
    Create tags
    :param db_connection:
        Session of the database connection
    :param tags:
        List of string type containing tags to create
    :return: [DBtag]
        The output will contain a list of tag objects
    """
    db_create_tags = []
    for tag_name in tags:
        db_create_tag = DBtag(name=tag_name)
        db_create_tags.append(db_create_tag)

    if db_create_tags:
        db_connection.add_all(db_create_tags)
        db_connection.flush()
        db_connection.commit()
    return db_create_tags


def get_rule_tag_names_by_rule_pack_version(db_connection: Session, rule_pack_version: str):
    """
    Get rule names and there tags based on the rule pack version
    :param db_connection:
        Session of the database connection
    :param rule_pack_version:
        Version of the rule pack for which to retrieve the rule tags
    :return: [rule.rule_name, tag.name]
        The output will contain a list of each rule and tag occurrence in the rule_pack
    """
    query: Query = db_connection.query(DBrule.rule_name, DBtag.name)
    query = query.join(DBruleTag, DBruleTag.tag_id == DBtag.id_)
    query = query.join(DBrule, DBrule.id_ == DBruleTag.rule_id)
    query = query.where(DBrule.rule_pack == rule_pack_version)
    return query.all()



# Standard Library
import logging

# Third Party
import requests

# First Party
from resc_backend.constants import (
    RWS_ROUTE_LAST_SCAN,
    RWS_ROUTE_REPOSITORIES,
    RWS_VERSION_PREFIX,
)
from resc_backend.resc_web_service.schema.repository import Repository

logger = logging.getLogger(__name__)


def create_repository(url: str, repository: Repository):
    api_url = f"{url}{RWS_VERSION_PREFIX}{RWS_ROUTE_REPOSITORIES}"
    response = requests.post(api_url, data=repository.json(), proxies={"http": "", "https": ""}, timeout=10)
    return response


def get_last_scan_for_repository(url: str, repository_id: int):
    api_url = f"{url}{RWS_VERSION_PREFIX}{RWS_ROUTE_REPOSITORIES}/{repository_id}{RWS_ROUTE_LAST_SCAN}"
    response = requests.get(api_url, proxies={"http": "", "https": ""}, timeout=10)
    return response



# Third Party
from pydantic import BaseModel, conint, conlist, constr

# First Party
from resc_backend.resc_web_service.schema.status_count import StatusCount


class RuleFindingCountModel(BaseModel):
    """
    :param Generic[Model]:
        Type of the object in the data list
    """

    rule_name: constr(max_length=100)
    finding_count: conint(gt=-1) = 0
    finding_statuses_count: conlist(item_type=StatusCount, min_items=None, max_items=500) = []

    class Config:
        orm_mode = True



# Standard Library
import datetime

# Third Party
from pydantic import BaseModel, HttpUrl, conint, constr

# First Party
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


class RepositoryEnrichedBase(BaseModel):
    project_key: constr(min_length=1, max_length=100)
    repository_id: constr(min_length=1, max_length=100)
    repository_name: constr(min_length=1, max_length=100)
    repository_url: HttpUrl
    vcs_provider: VCSProviders
    last_scan_id: conint(gt=0) = None
    last_scan_timestamp: datetime.datetime = None
    true_positive: conint(gt=-1)
    false_positive: conint(gt=-1)
    not_analyzed: conint(gt=-1)
    not_accessible: conint(gt=-1)
    clarification_required: conint(gt=-1)
    outdated: conint(gt=-1)
    total_findings_count: conint(gt=-1)


class RepositoryEnriched(RepositoryEnrichedBase):
    pass


class RepositoryEnrichedRead(RepositoryEnriched):
    id_: conint(gt=0)

    class Config:
        orm_mode = True



# First Party
from resc_backend.helpers.environment_wrapper import EnvironmentVariable

RESC_RABBITMQ_SERVICE_HOST = "RESC_RABBITMQ_SERVICE_HOST"
RESC_RABBITMQ_SERVICE_PORT_MGMT = "RESC_RABBITMQ_SERVICE_PORT_MGMT"
RABBITMQ_DEFAULT_VHOST = "RABBITMQ_DEFAULT_VHOST"
RABBITMQ_DEFAULT_USER = "RABBITMQ_DEFAULT_USER"
RABBITMQ_DEFAULT_PASSWORD = "RABBITMQ_DEFAULT_PASS"
RABBITMQ_USERNAME = "RABBITMQ_USERNAME"
RABBITMQ_PASSWORD = "RABBITMQ_PASSWORD"
DEBUG_MODE = "DEBUG_MODE"

REQUIRED_ENV_VARS = [
    EnvironmentVariable(
        RESC_RABBITMQ_SERVICE_HOST,
        "The hostname/IP address of the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        RESC_RABBITMQ_SERVICE_PORT_MGMT,
        "The port on which the rabbitmq mgmt server is running.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_DEFAULT_VHOST,
        "The virtual host name of the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_DEFAULT_USER,
        "The username used to connect to the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_DEFAULT_PASSWORD,
        "The password used to connect to the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_USERNAME,
        "The username used to connect to the rabbitmq queue",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_PASSWORD,
        "The password used to connect to the rabbitmq queue",
        required=True,
    ),
    EnvironmentVariable(
        DEBUG_MODE,
        "Show debug log statements, if set to '0' only INFO logs and above will be shown.",
        required=False,
        default="0",
    ),
]



# Third Party
from pydantic import BaseModel


class RescResponseModel(BaseModel):
    """
    Generic schema to be used for a response from resc scanner.
    """


class Model400(RescResponseModel):
    """
    Response schema to be used for a 400 BAD REQUEST.
    """

    class Config:
        schema_extra = {"example": {"detail": "Bad Request"}}


class Model404(RescResponseModel):
    """
    Response schema to be used for a 404 NOT FOUND.
    """

    class Config:
        schema_extra = {"example": {"data": {}, "detail": "<id> not found"}}


class Model409(RescResponseModel):
    """
    Response schema to be used for a 409 CONFLICT.
    """

    class Config:
        schema_extra = {"example": {"detail": "Unable to process entity due to conflict"}}


class Model422(RescResponseModel):
    """
    Response schema to be used for a 422 UNPROCESSABLE ENTITY.
    """

    class Config:
        schema_extra = {"example": {"detail": "Entity cannot be processed"}}



# First Party
from resc_backend.helpers.environment_wrapper import EnvironmentVariable

ENABLE_CORS = "ENABLE_CORS"
CORS_ALLOWED_DOMAINS = "CORS_ALLOWED_DOMAINS"

AUTHENTICATION_REQUIRED = "AUTHENTICATION_REQUIRED"
SSO_ACCESS_TOKEN_ISSUER_URL = "SSO_ACCESS_TOKEN_ISSUER_URL"
SSO_ACCESS_TOKEN_JWKS_URL = "SSO_ACCESS_TOKEN_JWKS_URL"
SSO_JWT_SIGN_ALGORITHM = "SSO_JWT_SIGN_ALGORITHM"
SSO_JWT_REQUIRED_CLAIMS = "SSO_JWT_REQUIRED_CLAIMS"
SSO_JWT_CLAIM_KEY_USER_ID = "SSO_JWT_CLAIM_KEY_USER_ID"
SSO_JWT_CLAIM_KEY_AUTHORIZATION = "SSO_JWT_CLAIM_KEY_AUTHORIZATION"
SSO_JWT_CLAIM_VALUE_AUTHORIZATION = "SSO_JWT_CLAIM_VALUE_AUTHORIZATION"

RESC_REDIS_CACHE_ENABLE = "RESC_REDIS_CACHE_ENABLE"
RESC_REDIS_SERVICE_HOST = "RESC_REDIS_SERVICE_HOST"
RESC_REDIS_SERVICE_PORT = "RESC_REDIS_SERVICE_PORT"
REDIS_PASSWORD = "REDIS_PASSWORD"

WEB_SERVICE_ENV_VARS = [
    EnvironmentVariable(
        ENABLE_CORS,
        "Enable by providing the value true to allow CORS requests.",
        required=False,
        default="",
    ),
    EnvironmentVariable(
        CORS_ALLOWED_DOMAINS,
        "Comma separated lists of domains to allow in the CORS policy if ENABLE_CORS is true",
        required=False,
        default="",
    ),
    EnvironmentVariable(
        AUTHENTICATION_REQUIRED,
        "set to false to disable authentication, any other value will enable SSO authentication",
        required=False,
        default="",
    ),
    EnvironmentVariable(
        RESC_REDIS_CACHE_ENABLE,
        "Set to true to enable the redis cache, its expected to be running separately from this instance",
        required=False,
        default="False",
    ),
]

CONDITIONAL_SSO_ENV_VARS = [
    EnvironmentVariable(
        SSO_ACCESS_TOKEN_ISSUER_URL,
        "URL of the access token issuer",
        required=True,
    ),
    EnvironmentVariable(
        SSO_ACCESS_TOKEN_JWKS_URL,
        "URL of the access token JWKS",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_SIGN_ALGORITHM,
        "signing algorithm used for the JWT, for example RS256",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_REQUIRED_CLAIMS,
        "Comma separated list of claims that need to be in the JWT",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_CLAIM_KEY_USER_ID,
        "JWT claim key used for the user id",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_CLAIM_KEY_AUTHORIZATION,
        "JWT claim key used for the authorization check",
        required=True,
    ),
    EnvironmentVariable(
        SSO_JWT_CLAIM_VALUE_AUTHORIZATION,
        "JWT claim value used for the authorization check. "
        "Used as string contains check on the key from SSO_JWT_CLAIM_KEY_AUTHORIZATION",
        required=True,
    ),
]

CONDITIONAL_REDIS_ENV_VARS = [
    EnvironmentVariable(
        RESC_REDIS_SERVICE_HOST,
        "The hostname/IP address of the REDIS server.",
        required=True,
    ),
    EnvironmentVariable(
        RESC_REDIS_SERVICE_PORT,
        "The port on which the REDIS server is running.",
        required=True,
    ),
    EnvironmentVariable(
        REDIS_PASSWORD,
        "The REDIS authentication secret.",
        required=True,
    ),
]



# Third Party
from pydantic import BaseModel, HttpUrl, conint, constr


class RepositoryBase(BaseModel):
    project_key: constr(min_length=1, max_length=100)
    repository_id: constr(min_length=1, max_length=100)
    repository_name: constr(min_length=1, max_length=100)
    repository_url: HttpUrl
    vcs_instance: conint(gt=0)


class Repository(RepositoryBase):
    pass


class RepositoryCreate(RepositoryBase):
    pass


class RepositoryRead(RepositoryBase):
    id_: conint(gt=0)

    class Config:
        orm_mode = True



# Third Party
from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint

# First Party
from resc_backend.db.model import Base
from resc_backend.resc_web_service.schema.repository import Repository


class DBrepository(Base):
    __tablename__ = "repository"
    id_ = Column("id", Integer, primary_key=True)
    vcs_instance = Column(Integer, ForeignKey("vcs_instance.id"), nullable=False)
    project_key = Column(String(100), nullable=False)
    repository_id = Column(String(100), nullable=False)
    repository_name = Column(String(100), nullable=False)
    repository_url = Column(String(200), nullable=False)
    __table_args__ = (
        UniqueConstraint(
            "project_key",
            "repository_id",
            "vcs_instance",
            name="unique_repository_id_per_project",
        ),
    )

    def __init__(self, project_key, repository_id, repository_name, repository_url, vcs_instance):
        self.project_key = project_key
        self.repository_id = repository_id
        self.repository_name = repository_name
        self.repository_url = repository_url
        self.vcs_instance = vcs_instance

    @staticmethod
    def create_from_repository(repository: Repository):
        db_repository = DBrepository(
            project_key=repository.project_key,
            repository_id=repository.repository_id,
            repository_name=repository.repository_name,
            repository_url=repository.repository_url,
            vcs_instance=repository.vcs_instance,
        )
        return db_repository



# Third Party
from sqlalchemy import Column, Enum, Integer, String, Text, UniqueConstraint

# First Party
from resc_backend.constants import AZURE_DEVOPS, BITBUCKET, GITHUB_PUBLIC
from resc_backend.db.model import Base


class DBVcsInstance(Base):
    __tablename__ = "vcs_instance"
    id_ = Column("id", Integer, primary_key=True)
    name = Column(String(200), nullable=False)
    provider_type = Column(Enum(BITBUCKET, AZURE_DEVOPS, GITHUB_PUBLIC, name="provider_type"))
    scheme = Column(String(50), nullable=False)
    hostname = Column(String(200), nullable=False)
    port = Column(Integer, nullable=False)
    organization = Column(String(200), nullable=True)
    scope = Column("vcs_scope", Text, nullable=True)
    exceptions = Column(Text, nullable=True)
    __table_args__ = (
        UniqueConstraint(
            "provider_type",
            "scheme",
            "hostname",
            "port",
            "organization",
            name="unique_vcs_instance",
        ),
    )

    def __init__(
        self,
        name: str,
        provider_type: str,
        scheme: str,
        hostname: str,
        port: int,
        organization: str,
        scope: str,
        exceptions: str,
    ):
        self.name = name
        self.provider_type = provider_type
        self.scheme = scheme
        self.hostname = hostname
        self.port = port
        self.organization = organization
        self.scope = scope
        self.exceptions = exceptions



# Third Party
from pydantic import BaseModel

# First Party
from resc_backend.resc_web_service.schema.finding_status import FindingStatus


class StatusCount(BaseModel):
    status: FindingStatus
    count: int = 0



# Standard Library
import datetime

# Third Party
from pydantic import BaseModel, HttpUrl, conint, constr, root_validator

# First Party
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


class DetailedFindingBase(BaseModel):
    file_path: str
    line_number: conint(gt=-1)
    column_start: conint(gt=-1)
    column_end: conint(gt=-1)
    commit_id: constr(max_length=120)
    commit_message: str
    commit_timestamp: datetime.datetime
    author: constr(max_length=200)
    email: constr(max_length=100)
    status: FindingStatus | None = FindingStatus.NOT_ANALYZED.value
    comment: constr(max_length=255) | None = None
    rule_name: constr(max_length=200)
    rule_pack: constr(max_length=100)
    project_key: constr(min_length=1, max_length=100)
    repository_name: constr(min_length=1, max_length=100)
    repository_url: HttpUrl
    timestamp: datetime.datetime
    vcs_provider: VCSProviders
    last_scanned_commit: constr(min_length=1, max_length=100)
    scan_id: conint(gt=0)
    event_sent_on: datetime.datetime | None


class DetailedFinding(DetailedFindingBase):
    pass


class DetailedFindingRead(DetailedFinding):
    id_: conint(gt=0)
    commit_url: constr(min_length=1) | None

    @staticmethod
    def build_bitbucket_commit_url(
        repository_url: str,
        repository_name: str,
        project_key: str,
        file_path: str,
        commit_id: str,
    ) -> str:
        arr = repository_url.split("/")
        if len(arr) >= 3:
            repo_base_url = arr[0] + "//" + arr[2]
        else:
            repo_base_url = repository_url
        bitbucket_commit_url = (
            f"{repo_base_url}/projects/{project_key}/repos/" f"{repository_name}/browse/{file_path}?at={commit_id}"
        )
        commit_url = bitbucket_commit_url
        return commit_url

    @staticmethod
    def build_ado_commit_url(repository_url: str, file_path: str, commit_id: str) -> str:
        ado_commit_url = f"{repository_url}/commit/{commit_id}?path=/{file_path}"
        return ado_commit_url

    @staticmethod
    def build_github_commit_url(repository_url: str, file_path: str, commit_id: str) -> str:
        github_commit_url = f"{repository_url}/commit/{commit_id}?path=/{file_path}"
        return github_commit_url

    @root_validator
    def build_commit_url(cls, values) -> dict:  # noqa: N805
        if values["status"] is None:
            values["status"] = FindingStatus.NOT_ANALYZED.value
        if values["comment"] is None:
            values["comment"] = ""
        if values["vcs_provider"] == VCSProviders.BITBUCKET:
            values["commit_url"] = cls.build_bitbucket_commit_url(
                repository_url=values["repository_url"],
                repository_name=values["repository_name"],
                project_key=values["project_key"],
                file_path=values["file_path"],
                commit_id=values["commit_id"],
            )
        elif values["vcs_provider"] == VCSProviders.AZURE_DEVOPS:
            values["commit_url"] = cls.build_ado_commit_url(
                repository_url=values["repository_url"],
                file_path=values["file_path"],
                commit_id=values["commit_id"],
            )

        elif values["vcs_provider"] == VCSProviders.GITHUB_PUBLIC:
            values["commit_url"] = cls.build_github_commit_url(
                repository_url=values["repository_url"],
                file_path=values["file_path"],
                commit_id=values["commit_id"],
            )
        else:
            raise NotImplementedError(f"Unsupported VCSProvider: {values['vcs_provider']}")
        return values

    class Config:
        orm_mode = True
